import os
from datetime import datetime

from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt, pyqtSignal
from opentps.core.data import Patient
from opentps.core.data.plan import RTPlan
from opentps.core.processing.doseCalculation.protons.mcsquareDoseCalculator import MCsquareDoseCalculator
from opentps.core.processing.planEvaluation.robustnessEvaluation import RobustnessEvalProton
from opentps.gui.panels.patientDataWidgets import PatientDataComboBox
from opentps.gui.panels.planDesignPanel.robustnessSettings import RobustnessSettings


class PlanEvaluationPanel(QWidget):
    Robustness_analysis_recomputed = pyqtSignal()

    def __init__(self, viewController):
        QWidget.__init__(self)

        self._patient:Patient = None
        self._viewController = viewController

        self.robustness_scenarios = []

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self._planLabel = QLabel('plan:')
        self.layout.addWidget(self._planLabel)
        self._planComboBox = PatientDataComboBox(patientDataType=RTPlan, patient=self._patient, parent=self)
        self.layout.addWidget(self._planComboBox)

        self._robustSettings = RobustnessSettings(self._viewController, planEvaluation=True, parent=self)
        self.layout.addWidget(self._robustSettings)

        self.layout.addSpacing(30)
        self.button_hLoayout = QHBoxLayout()
        self.layout.addLayout(self.button_hLoayout)
        self.ComputeScenariosButton = QPushButton('Compute \n Scenarios')
        self.ComputeScenariosButton.setEnabled(False)
        self.button_hLoayout.addWidget(self.ComputeScenariosButton)
        self.ComputeScenariosButton.clicked.connect(self.computeRobustnessScenarios)
        self.layout.addWidget(QLabel('<b>!!!Disabled (refactoring)!!!</b>'))
        self.layout.addWidget(QLabel('<b>!!!Plan evaluation soon available!!!</b>'))
        self.layout.addSpacing(40)

        self.layout.addWidget(QLabel('<b>Displayed dose:</b>'))
        self.DisplayedDose = QComboBox()
        self.DisplayedDose.addItems(['Nominal', 'Worst scenario', 'Voxel wise minimum', 'Voxel wise maximum'])
        self.layout.addWidget(self.DisplayedDose)
        self.layout.addSpacing(30)

        self.layout.addWidget(QLabel('<b>Prescription:</b>'))
        self.Prescription = QDoubleSpinBox()
        self.Prescription.setRange(0.0, 100.0)
        self.Prescription.setSingleStep(1.0)
        self.Prescription.setValue(60.0)
        self.Prescription.setSuffix(" Gy")
        self.layout.addWidget(self.Prescription)
        self.layout.addSpacing(10)

        self.layout.addWidget(QLabel('<b>Evaluation metric:</b>'))
        self.Metric = QComboBox()
        self.Metric.addItems(['D95', 'MSE'])
        self.layout.addWidget(self.Metric)
        self.layout.addSpacing(10)

        self.CI_label = QLabel('<b>Confidence interval:</b>')
        self.layout.addWidget(self.CI_label)
        self.CI = QSlider(Qt.Horizontal)
        self.CI.setMinimum(0)
        self.CI.setMaximum(100)
        self.CI.setValue(90)
        self.layout.addWidget(self.CI)
        self.layout.addStretch()

        self.Metric.currentIndexChanged.connect(self.recompute_robustness_analysis)
        self.DisplayedDose.currentIndexChanged.connect(self.recompute_robustness_analysis)
        self.Prescription.valueChanged.connect(self.recompute_robustness_analysis)

        self.setCurrentPatient(self._viewController.currentPatient)
        self._viewController.currentPatientChangedSignal.connect(self.setCurrentPatient)

        self.recompute_robustness_analysis()

    def setCurrentPatient(self, patient:Patient):
        self._planComboBox.setPatient(patient)

    @property
    def selectedPlan(self):
        return self._planComboBox.selectedPlan

    def computeRobustnessScenarios(self):
        # TODO: Take CT, target, etc. from plan.planDesign. Same for MC2 config

        # find selected CT image
        if (self.CT_disp_ID < 0):
            print("Error: No CT image selected")
            return
        ct_patient_id, ct_id = self.Patients.find_CT_image(self.CT_disp_ID)
        ct = self.Patients.list[ct_patient_id].CTimages[ct_id]

        # find selected plan
        if (self.Plan_disp_ID < 0):
            print("Error: No treatment plan selected")
            return
        plan_patient_id, plan_id = self.Patients.find_plan(self.Plan_disp_ID)
        plan = self.Patients.list[plan_patient_id].Plans[plan_id]

        # find contours
        Target_name = self.Target.currentText()
        patient_id, struct_id, contour_id = self.Patients.find_contour(Target_name)
        AllContours = self.Patients.list[patient_id].RTstructs[struct_id].Contours

        # configure MCsquare module
        doseCalculator = MCsquareDoseCalculator()
        doseCalculator.BDL.selected_BDL = self.Dose_calculation_param["BDL"]
        doseCalculator.Scanner.selected_Scanner = self.Dose_calculation_param["Scanner"]
        doseCalculator.NumProtons = self.Dose_calculation_param["NumProtons"]
        doseCalculator.MaxUncertainty = self.Dose_calculation_param["MaxUncertainty"]
        doseCalculator.dose2water = self.Dose_calculation_param["dose2water"]
        doseCalculator.SetupSystematicError = self.RobustEval["syst_setup"]
        doseCalculator.SetupRandomError = self.RobustEval["rand_setup"]
        doseCalculator.RangeSystematicError = self.RobustEval["syst_range"]

        if (self.RobustEval["Strategy"] == 'DoseSpace'):
            doseCalculator.Robustness_Strategy = "DoseSpace"
        elif (self.RobustEval["Strategy"] == 'ErrorSpace_stat'):
            doseCalculator.Robustness_Strategy = "ErrorSpace_stat"
        else:
            doseCalculator.Robustness_Strategy = "ErrorSpace_regular"

        # Crop CT image with contour:
        if (self.Dose_calculation_param["CropContour"] == "None"):
            doseCalculator.Crop_CT_contour = {}
        else:
            patient_id, struct_id, contour_id = self.Patients.find_contour(self.Dose_calculation_param["CropContour"])
            doseCalculator.Crop_CT_contour = self.Patients.list[patient_id].RTstructs[struct_id].Contours[contour_id]

        # run MCsquare simulation
        scenarios = doseCalculator.MCsquare_RobustScenario_calculation(ct, plan, AllContours)

        # save data
        output_path = os.path.join(self.data_path, "OpenTPS")
        if not os.path.isdir(output_path):
            os.mkdir(output_path)
        output_folder = os.path.join(output_path,
                                     "RobustnessTest_" + datetime.datetime.today().strftime("%b-%d-%Y_%H-%M-%S"))
        scenarios.save(output_folder)

        self.robustness_scenarios = scenarios
        self.recompute_robustness_analysis()

    def recompute_robustness_analysis(self):
        # TODO: Take CT, target, etc. from plan.planDesign. Same for MC2 config

        if (self.robustness_scenarios == []): return

        CI = self.CI.value()
        self.CI_label.setText('<b>Confidence interval:</b> &nbsp;&nbsp;&nbsp; ' + str(CI) + " %")

        # target contour
        Target_name = self.Target.currentText()
        if (Target_name == ''): return
        patient_id, struct_id, contour_id = self.Patients.find_contour(Target_name)
        Target = self.Patients.list[patient_id].RTstructs[struct_id].Contours[contour_id]
        TargetPrescription = self.Prescription.value()

        self.robustness_scenarios.DoseDistributionType = self.DisplayedDose.currentText()
        # TODO: update with new attribut robustnessEval.analysisStrategy.DOSIMETRIC
        if (self.robustness_scenarios.SelectionStrategy == RobustnessEvalProton.Strategies.DOSIMETRIC):
            self.robustness_scenarios.dosimetric_space_analysis(self.Metric.currentText(), CI, Target,
                                                                TargetPrescription)
        else:
            self.robustness_scenarios.error_space_analysis(self.Metric.currentText(), Target, TargetPrescription)

        self.Robustness_analysis_recomputed.emit()


import os

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QPushButton, QFrame, QMessageBox, QFileDialog, QLabel, QHBoxLayout

from opentps.core import APIInterpreter
from opentps.gui.panels.scriptingPanel.scriptingWindow import ScriptingWindow


class ScriptingPanel(QWidget):
    newScriptingWindowSignal = pyqtSignal()

    def __init__(self):
        QWidget.__init__(self)

        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

        self._newScriptingWindowButton = QPushButton('New scripting window')
        self._newScriptingWindowButton.clicked.connect(self.newScriptingWindow)
        self._layout.addWidget(self._newScriptingWindowButton)

        self._newScriptFileViewButton = QPushButton('Select new script file')
        self._newScriptFileViewButton.clicked.connect(self.newScriptFile)
        self._layout.addWidget(self._newScriptFileViewButton)

        self._scriptFileViewsFrame = QFrame()
        self._scriptFileViewsFrame.setFrameShape(QFrame.StyledPanel)
        self._layout.addWidget(self._scriptFileViewsFrame)

        self._scriptFileViewsLayout = QVBoxLayout()
        self._scriptFileViewsLayout.setContentsMargins(0, 0, 0, 0)
        self._scriptFileViewsFrame.setLayout(self._scriptFileViewsLayout)

        self._layout.addStretch()

    def newScriptingWindow(self):
        self.scriptingWindow = ScriptingWindow()
        self.scriptingWindow.show()

    def newScriptFile(self):
        self._scriptFileViewsLayout.addWidget(ScriptingFileView())


class ScriptingFileView(QWidget):
    _scriptPath = None

    def __init__(self):
        QWidget.__init__(self)

        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

        self._topFrame = QFrame()
        self._topFrame.setFrameShape(QFrame.StyledPanel)
        self._layout.addWidget(self._topFrame)

        self._bottomFrame = QFrame()
        self._bottomFrame.setFrameShape(QFrame.StyledPanel)
        self._layout.addWidget(self._bottomFrame)

        topLayout = QHBoxLayout()
        topLayout.setContentsMargins(0, 0, 0, 0)
        self._topFrame.setLayout(topLayout)

        bottomLayout = QHBoxLayout()
        bottomLayout.setContentsMargins(0, 0, 0, 0)
        self._bottomFrame.setLayout(bottomLayout)

        closeButton = QPushButton('X')
        closeButton.clicked.connect(self.close)

        selectFileButton = QPushButton('Select file')
        selectFileButton.clicked.connect(self._selectAndSetFile)

        runButton = QPushButton('Run')
        runButton.clicked.connect(self._runFile)

        topLayout.addWidget(closeButton)
        topLayout.addWidget(selectFileButton)
        topLayout.addWidget(runButton)

        self._fileNameLabel = QLabel()
        bottomLayout.addWidget(self._fileNameLabel)

    def close(self):
        self.setParent(None)

    def _selectAndSetFile(self):
        self._scriptPath = self._openFileSelection()
        self._setFileNameLabel()

    def _openFileSelection(self):
        filePath, _ = QFileDialog.getOpenFileName(self, "Select a python script", self._scriptPath, "python script (*.py)")
        return filePath

    def _setFileNameLabel(self):
        self._fileNameLabel.setText(os.path.basename(self._scriptPath))

    def _runFile(self):
        msg = QMessageBox()
        msg.setWindowTitle(os.path.basename(self._scriptPath))
        msg.setIcon(QMessageBox.Information)

        code = self._readFile()

        try:
            output = APIInterpreter.run(code)
            msg.setText(output)
        except Exception as err:
            msg.setText(format(err))
            raise err from err

        msg.exec_()

    def _readFile(self):
        code = ''

        with open(self._scriptPath, 'r') as file:
            code = file.read()

        return code



import os

from PyQt5.QtCore import QSize
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QFileDialog
from pyqtgraph import PlotWidget, PlotCurveItem, mkPen
from pyqtgraph.exporters import ImageExporter

import opentps.gui.res.icons as IconModule

class ProfileViewer(QWidget):
    def __init__(self, viewController, nbProfiles=10):
        QWidget.__init__(self)

        self._layout = QHBoxLayout(self)
        self._toolbar = _ProfileToolbar(self, viewController)
        self._viewController = viewController

        self._profilePlot = _ProfilePlot()
        self._profiles = []

        for i in range(nbProfiles):
            self._profiles.append(self._profilePlot.newProfile([0, 0], [0, 0]))

        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)
        self._layout.addWidget(self._toolbar)
        self._layout.addWidget(self._profilePlot)

    def count(self):
        return len(self._profiles)

    def drawProfile(self, profileIndex, *args, **kwargs):
        axY = self._profilePlot.getAxis('left')
        yRange = axY.range

        self.removeProfile(profileIndex)
        self._profiles[profileIndex] = self._profilePlot.newProfile(*args, **kwargs)

        if profileIndex>2:
            self._profilePlot.setYRange(yRange[0], yRange[1], padding=0.)

    def removeProfile(self, profileIndex):
        profile = self._profiles[profileIndex]
        self._profilePlot.removeItem(profile)

    @property
    def nbProfiles(self):
        return len(self._profiles)

    def erase(self):
        for profile in self._profiles:
            profile.setData([0, 0], [0, 0])

    def export(self):
        self._profilePlot.export()


class _ProfilePlot(PlotWidget):
    def __init__(self):
        PlotWidget.__init__(self)

        self.addLegend()
        self.getPlotItem().setContentsMargins(5, 0, 20, 5)
        self.setBackground('k')
        self.setTitle("Profiles")
        self.setLabel('left', 'Intensity')
        self.setLabel('bottom', 'Distance (mm)')

    def export(self):
        exporter = ImageExporter(self.getPlotItem())

        fileName = self.fileDialogForExporter(exporter)

        if fileName is None:
            return

        exporter.export(fileName[0])

    def fileDialogForExporter(self, exporter):
        dlg = QFileDialog()
        dlg.setAcceptMode(QFileDialog.AcceptSave)
        dlg.setFileMode(QFileDialog.AnyFile)

        supportedExtensions = exporter.getSupportedImageFormats()
        str = ''
        for extension in supportedExtensions:
            str += extension + ' (' + extension + ')' + ';;'
        str = str[:-2]

        fileName = dlg.getSaveFileName(filter=str)
        return fileName

    def newProfile(self, *args, **kwargs) -> PlotCurveItem:
        pl = PlotCurveItem(*args, **kwargs)
        self.addItem(pl)

        return pl

class _ProfileToolbar(QWidget):
    def __init__(self, profileViewer, viewController):
        QWidget.__init__(self)

        self._profileViewer = profileViewer
        self._viewController = viewController
        self._layout = QVBoxLayout(self)

        self.setLayout(self._layout)

        iconPath = IconModule.__path__[0] + os.path.sep

        icon = QIcon(iconPath+"pencil--plus.png")
        self._buttonNewProfile = QPushButton()
        self._buttonNewProfile.setIcon(icon)
        self._buttonNewProfile.setIconSize(QSize(16, 16))
        self._buttonNewProfile.clicked.connect(self._setProfileWidgetEnabled)

        icon = QIcon(iconPath + "cross.png")
        self._buttonStop = QPushButton()
        self._buttonStop.setIcon(icon)
        self._buttonStop.setIconSize(QSize(16, 16))
        self._buttonStop.clicked.connect(self._setProfileWidgetDisabled)

        icon = QIcon(iconPath + "disk.png")
        self._buttonSave = QPushButton()
        self._buttonSave.setIcon(icon)
        self._buttonSave.setIconSize(QSize(16, 16))
        self._buttonSave.clicked.connect(self._profileViewer.export)

        self.setLayout(self._layout)

        self.setAutoFillBackground(True)
        self._layout.addWidget(self._buttonNewProfile)
        self._layout.addWidget(self._buttonStop)
        self._layout.addWidget(self._buttonSave)

        self._layout.addStretch(1)

    def _setProfileWidgetEnabled(self):
        self._viewController.profileWidgetEnabled = True
        self._viewController.profileWidgetCallback.setPrimaryImageData = \
            lambda *args, **kwargs: self._drawImageProfile(0, *args, **kwargs, pen=mkPen(width=1, color=[125, 125, 125]))
        self._viewController.profileWidgetCallback.setSecondaryImageData = \
            lambda *args, **kwargs: self._drawImageProfile(1, *args, **kwargs, pen=mkPen(width=1, color=[255, 255, 255]))
        self._viewController.profileWidgetCallback.setContourData = \
            lambda *args, **kwargs: self._drawContourProfiles(*args, **kwargs)

    def _drawImageProfile(self, ind, *args, name='', **kwargs):
        self._profileViewer.drawProfile(ind, *args, **kwargs, name=name)

    def _drawContourProfiles(self, contourData, name=[], pen=[]):
        currentNbProfiles = self._profileViewer.count()
        for i in range(currentNbProfiles):
            if currentNbProfiles-i-1>1:
                self._profileViewer.drawProfile(currentNbProfiles-i-1, [0, 0], [0, 0])

        for i, n in enumerate(name):
            x, y = contourData[i]
            self._profileViewer.drawProfile(2+i, x, y, name=n, pen=pen[i])

    def _setProfileWidgetDisabled(self):
        self._viewController.profileWidgetEnabled = False
        self._viewController.profileWidgetCallback.setPrimaryImageData = lambda *args, **kwargs: None
        self._viewController.profileWidgetCallback.setSecondaryImageData = lambda *args, **kwargs: None
        self._viewController.profileWidgetCallback.setContourData = lambda *args, **kwargs: None



from functools import partial
from typing import Sequence, Union, Optional

import vtkmodules.vtkRenderingCore as vtkRenderingCore

from opentps.core.data import ROIContour
from opentps.core.data.images import ROIMask, Image3D
from opentps.gui.viewer.dataForViewer.ROIContourForViewer import ROIContourForViewer
from opentps.gui.viewer.dataForViewer.ROIMaskForViewer import ROIMaskForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.lookupTables import uniqueColorLTTo3DLT
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.secondaryImage3DLayer_3D import \
    SecondaryImage3DLayer_3D


class ContourLayer_3D:
    def __init__(self, renderer, renderWindow, iStyle):
        self._contours = [] # Acts as a cache
        self._referenceImage = None
        self._renderer = renderer
        self._renderWindow = renderWindow
        self._iStyle = iStyle
        self._vtkContours = []
        self._partialHandlers = []

    def close(self):
        for vtkContour in self._vtkContours:
            vtkContour.close()

    @property
    def contours(self) -> Sequence[Union[ROIContourForViewer, ROIMaskForViewer]]:
        return [contour for contour in self._contours]

    def setNewContour(self, contour:Union[ROIContour, ROIMask]):
        if isinstance(contour, ROIContour):
            contour = ROIContourForViewer(contour)
        elif isinstance(contour, ROIMask):
            contour = ROIMaskForViewer(contour)
        else:
            raise ValueError(str(type(contour)) + ' is not a valid type for a contour.')

        if contour in self._contours:
            return

        if not contour.visible:
            return

        #contour.referenceImage = self.referenceImage

        self._contours.append(contour)

        vtkContourObj = MaskLayer_3D(self._renderer, self._renderWindow, self._iStyle)

        if isinstance(contour, ROIContourForViewer):
            vtkContourObj.image = contour.asROIMaskForViewer()
        else:
            vtkContourObj.image = contour

        self._vtkContours.append(vtkContourObj)

        partialHandler = partial(self._handleVisibilityChange, contour)
        self._partialHandlers.append(partialHandler)
        contour.visibleChangedSignal.connect(partialHandler)

        self._renderWindow.Render()

    def update(self):
        for vtkContour in self._vtkContours:
            vtkContour.update()

    def _handleVisibilityChange(self, contour:Union[ROIContourForViewer, ROIMaskForViewer], visible):
        if not contour.visible:
            self._removeContour(contour)

    def _removeContour(self, contour:Union[ROIContourForViewer, ROIMaskForViewer]):
        contourIndex = self._contours.index(contour)
        vtkContourObj = self._vtkContours[contourIndex]
        partialHandler = self._partialHandlers[contourIndex]

        vtkContourObj.close()

        self._contours.remove(contour)
        self._vtkContours.remove(vtkContourObj)
        self._partialHandlers.remove(partialHandler)

        contour.visibleChangedSignal.disconnect(partialHandler)

    @property
    def referenceImage(self) -> Optional[Image3D]:
        return self._referenceImage

    @referenceImage.setter
    def referenceImage(self, image: Image3D):
        self._referenceImage = image

        for contour in self._contours:
            contour.referenceImage = self._referenceImage


class MaskLayer_3D(SecondaryImage3DLayer_3D):
    def __init__(self, renderer, renderWindow, iStyle):
        super().__init__(renderer, renderWindow, iStyle)

        self._volumeProperty = vtkRenderingCore.vtkVolumeProperty()
        self._volumeProperty.SetInterpolationTypeToLinear()
        self._volumeProperty.ShadeOn()
        self._volumeProperty.SetAmbient(0.4)
        self._volumeProperty.SetDiffuse(0.6)
        self._volumeProperty.SetSpecular(0.2)

    def _updateLookupTable(self, lt):
        volumeColor, volumeScalarOpacity, volumeGradientOpacity = uniqueColorLTTo3DLT(lt)
        self._volumeProperty.SetColor(volumeColor)
        self._volumeProperty.SetScalarOpacity(volumeScalarOpacity)
        self._volumeProperty.SetGradientOpacity(volumeGradientOpacity)

        self._renderWindow.Render()




from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer


class PolyDataForViewer(GenericImageForViewer):

    def __init__(self, vtkPolyData):
        super().__init__(vtkPolyData)

        if hasattr(self, 'dataChangedSignal'):
            return
        
        self.dataChangedSignal = Event()

    @property
    def vtkOutputPort(self):
        return self.GetOutputPort()



from PyQt5.QtWidgets import *


class BeamDialog(QDialog):
    def __init__(self, Modality, BeamName, RangeShifterList=[]):
        RangeShifterList = ["None"] + RangeShifterList

        # initialize the window
        QDialog.__init__(self)
        self.setWindowTitle('Add beam')
        self.resize(300, 150)
        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)

        # form
        self.InputLayout = QGridLayout()
        self.main_layout.addLayout(self.InputLayout)

        self.InputLayout.addWidget(QLabel('<b>Name:</b>'), 0, 0)
        self.BeamName = QLineEdit(BeamName)
        self.InputLayout.addWidget(self.BeamName, 0, 1)

        self.InputLayout.addWidget(QLabel('<b>Gantry angle:</b>'), 1, 0)
        self.GantryAngle = QDoubleSpinBox()
        self.GantryAngle.setRange(0.0, 360.0)
        self.GantryAngle.setSingleStep(5.0)
        self.GantryAngle.setValue(0.0)
        self.GantryAngle.setSuffix("°")
        self.InputLayout.addWidget(self.GantryAngle, 1, 1)

        self.InputLayout.addWidget(QLabel('<b>Couch angle:</b>'), 2, 0)
        self.CouchAngle = QDoubleSpinBox()
        self.CouchAngle.setRange(0.0, 360.0)
        self.CouchAngle.setSingleStep(5.0)
        self.CouchAngle.setValue(0.0)
        self.CouchAngle.setSuffix("°")
        self.InputLayout.addWidget(self.CouchAngle, 2, 1)

        self.InputLayout.addWidget(QLabel('<b>Range shifter:</b>'), 3, 0)
        self.RangeShifter = QComboBox()
        self.RangeShifter.addItems(RangeShifterList)
        self.InputLayout.addWidget(self.RangeShifter, 3, 1)
        if Modality == "IMRT":
            self.RangeShifter.setEnabled(False)

        # buttons
        self.ButtonLayout = QHBoxLayout()
        self.main_layout.addLayout(self.ButtonLayout)
        self.CancelButton = QPushButton('Cancel')
        self.ButtonLayout.addWidget(self.CancelButton)
        self.CancelButton.clicked.connect(self.reject)
        self.AddButton = QPushButton('Add')
        self.AddButton.clicked.connect(self.accept)
        self.ButtonLayout.addWidget(self.AddButton)



import copy
import logging
import numpy as np
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QComboBox, QLabel, QPushButton, QCheckBox, QDialog, \
    QLineEdit, QDoubleSpinBox, QHBoxLayout
from PyQt5.QtCore import Qt
from opentps.core.data.images import CTImage
from opentps.core.data.plan import ObjectivesList
from opentps.core.data.plan import ProtonPlanDesign,Robustness
from opentps.core.data._patient import Patient
from opentps.core.data.plan._photonPlanDesign import PhotonPlanDesign
from opentps.core.data.plan._rtPlanDesign import RTPlanDesign
from opentps.core.io import mcsquareIO
from opentps.core.io.scannerReader import readScanner
from opentps.core.processing.doseCalculation.doseCalculationConfig import DoseCalculationConfig
from opentps.core.processing.doseCalculation.photons.cccDoseCalculator import CCCDoseCalculator
from opentps.core.processing.doseCalculation.protons.mcsquareDoseCalculator import MCsquareDoseCalculator
from opentps.core.processing.planOptimization.planOptimization import BoundConstraintsOptimizer, IntensityModulationOptimizer
from opentps.gui.panels.doseComputationPanel import DoseComputationPanel
from opentps.gui.panels.patientDataWidgets import PatientDataComboBox
from opentps.gui.panels.planOptimizationPanel.objectivesWindow import ObjectivesWindow
from opentps.gui.panels.planOptimizationPanel.optimizationSettings import OptiSettingsDialog

logger = logging.getLogger(__name__)



class doseCalculationWindow(QDialog):
    def __init__(self, viewController, radiationType, parent=None, contours=None, beamlets=True, robustOpti=False):
        super().__init__(parent)

        self._viewController = viewController
        self._contours = contours
        self._beamlets = beamlets
        self._robustOpti = robustOpti
        self._radiationType = radiationType

        self._doseComputationPanel = DoseComputationPanel(viewController)

        self._doseComputationPanel._runButton.hide()
        self._doseComputationPanel._mcsquareConfigWidget.hide()

        if self._beamlets:
            if self._radiationType == "PROTON":
                self._doseComputationPanel._doseSpacingLabel.show()
                self._doseComputationPanel._numProtons.setValue(int(5e4))
            self._doseComputationPanel._cropBLBox.show()
            if not(self._doseComputationPanel._cropBLBox.isChecked()):
                self._contours = None

            self._beamletButton = QPushButton('Compute beamlets')
            self._beamletButton.clicked.connect(self._computeBeamlets)
            self._doseComputationPanel.layout.addWidget(self._beamletButton)
        else:
            self._optimizeButton = QPushButton('Beamlet-free optimize')
            self._optimizeButton.clicked.connect(self._optimizeBeamletFree)
            self._doseComputationPanel.layout.addWidget(self._optimizeButton)

        self.layout = QVBoxLayout(self)
        self.setLayout(self.layout)
        self.layout.addWidget(self._doseComputationPanel)

    def setCT(self, ct):
        self._doseComputationPanel.selectedCT = ct

    def setPlan(self, plan):
        self._doseComputationPanel.selectedPlan = plan

    def _computeBeamlets(self):
        settings = DoseCalculationConfig()
        calibration = readScanner(settings.scannerFolder)
        if self._radiationType == "PROTON":
            doseCalculator = MCsquareDoseCalculator()
            beamModel = mcsquareIO.readBDL(settings.bdlFile)
            doseCalculator.beamModel = beamModel
            if self._doseComputationPanel._doseSpacingLabel.isChecked():
                self._doseComputationPanel.selectedPlan.planDesign.setScoringParameters(scoringSpacing=self._doseComputationPanel._doseSpacingSpin.value(), adapt_gridSize_to_new_spacing=True)
                # self._doseComputationPanel.selectedPlan.planDesign.scoringVoxelSpacing = self._doseComputationPanel._doseSpacingSpin.value()
            doseCalculator.nbPrimaries = self._doseComputationPanel._numProtons.value()
            doseCalculator.statUncertainty = self._doseComputationPanel._statUncertainty.value()
            doseCalculator.ctCalibration = calibration
            doseCalculator.overwriteOutsideROI = self._doseComputationPanel._selectedROI

        elif self._radiationType == "PHOTON":
            doseCalculator = CCCDoseCalculator(batchSize=self._doseComputationPanel._batchSize.value())
            doseCalculator.ctCalibration = calibration
            doseCalculator.overwriteOutsideROI = self._doseComputationPanel._selectedROI
        else: logger.error("Could not identify plan radiation type")
    
        if self._robustOpti:
            nominal, scenarios = doseCalculator.computeRobustScenarioBeamlets(self._doseComputationPanel.selectedCT,
                                                self._doseComputationPanel.selectedPlan, self._contours)
            self._doseComputationPanel.selectedPlan.planDesign.beamlets = nominal
            self._doseComputationPanel.selectedPlan.planDesign.robustness.scenarios = scenarios
            self._doseComputationPanel.selectedPlan.planDesign.robustness.numScenarios = len(scenarios)
        else:
            beamlets = doseCalculator.computeBeamlets(self._doseComputationPanel.selectedCT,
                                                self._doseComputationPanel.selectedPlan, self._contours)
            self._doseComputationPanel.selectedPlan.planDesign.beamlets = beamlets    

        self.accept()

    def _optimizeBeamletFree(self):
        settings = DoseCalculationConfig()

        beamModel = mcsquareIO.readBDL(settings.bdlFile)
        calibration = readScanner(settings.scannerFolder)

        doseCalculator = MCsquareDoseCalculator()
        doseCalculator.beamModel = beamModel
        doseCalculator.nbPrimaries = self._doseComputationPanel._numProtons.value()
        doseCalculator.ctCalibration = calibration

        doseImage = doseCalculator.optimizeBeamletFree(self._doseComputationPanel.selectedCT,
                                                       self._doseComputationPanel.selectedPlan,
                                                       self._contours)
        doseImage.patient = self._doseComputationPanel.selectedCT.patient
        self.accept()

class PlanOptiPanel(QWidget):
    _optiAlgos = ["Scipy-LBFGS (recommended)", "Scipy-BFGS", "In-house Gradient", "In-house LBFGS", "In-house BFGS", "FISTA",
                  "LP", "Beamlet-free MCsquare"]

    def __init__(self, viewController):
        QWidget.__init__(self)

        self._patient: Patient = None
        self._radiationType = "PROTON" # default
        self._optiConfig = {"method": "Scipy-LBFGS", "maxIter": 1000, "step": 0.02, "bounds": None}

        self._viewController = viewController

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        # Add the radiation type label
        self._radiationLabel = QLabel(self)
        self._radiationLabel.setFixedSize(100, 30)  
        self._radiationLabel.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(self._radiationLabel)

        self._planStructureLabel = QLabel('Plan design:')
        self.layout.addWidget(self._planStructureLabel)
        self._planStructureComboBox = PatientDataComboBox(patientDataType=RTPlanDesign, patient=self._patient,
                                                          parent=self)
        self._planStructureComboBox.selectedDataEvent.connect(self._handlePlanStructure)
        self.layout.addWidget(self._planStructureComboBox)

        self._ctLabel = QLabel('CT:')
        self.layout.addWidget(self._ctLabel)
        self._ctComboBox = PatientDataComboBox(patientDataType=CTImage, patient=self._patient, parent=self)
        self.layout.addWidget(self._ctComboBox)

        self._objectivesWidget = ObjectivesWidget(self._viewController)
        self._objectivesWidget.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self._objectivesWidget)

        self.layout.addWidget(QLabel('Optimization algorithm:'))
        self._algoBox = QComboBox()
        self._algoBox.addItem(self._optiAlgos[0])
        self._algoBox.addItem(self._optiAlgos[1])
        self._algoBox.addItem(self._optiAlgos[2])
        self._algoBox.addItem(self._optiAlgos[3])
        self._algoBox.addItem(self._optiAlgos[4])
        self._algoBox.addItem(self._optiAlgos[5])
        self._algoBox.addItem(self._optiAlgos[6])
        self._algoBox.addItem(self._optiAlgos[7])
        self._algoBox.currentIndexChanged.connect(self._handleAlgo)
        self.layout.addWidget(self._algoBox)

        self._configButton = QPushButton('Advanced configuration')
        self._configButton.clicked.connect(self._openConfig)
        self.layout.addWidget(self._configButton)

        self._beamletGridBox = QCheckBox('Generate beamlet grid')
        self._beamletGridBox.setChecked(True)
        self.layout.addWidget(self._beamletGridBox)
        self._spotPlacementBox = QCheckBox('Place spots')
        self._spotPlacementBox.setChecked(True)
        self._beamletBox = QCheckBox('Compute beamlets')
        self._beamletBox.setChecked(True)
        self.layout.addWidget(self._spotPlacementBox)
        self.layout.addWidget(self._beamletBox)

        self._fractionsLayout = QHBoxLayout()
        self.layout.addLayout(self._fractionsLayout)

        self._fractionsLabel = QLabel('Number of fractions:')
        self._fractionsLayout.addWidget(self._fractionsLabel)
        self._fractionsSpin = QDoubleSpinBox()
        self._fractionsSpin.setRange(1, 99)
        self._fractionsSpin.setValue(1)
        self._fractionsSpin.setDecimals(0)
        self._fractionsLayout.addWidget(self._fractionsSpin)

        self._planLabel = QLabel('plan name:')
        self.layout.addWidget(self._planLabel)
        self._planNameEdit = QLineEdit(self)
        self._planNameEdit.setText('New plan')
        self.layout.addWidget(self._planNameEdit)

        self._runButton = QPushButton('Optimize plan')
        self._runButton.clicked.connect(self._run)
        self.layout.addWidget(self._runButton)

        self.layout.addStretch()

        self.setCurrentPatient(self._viewController.currentPatient)
        self._viewController.currentPatientChangedSignal.connect(self.setCurrentPatient)

        self._handleAlgo()
        #self._handlePlanStructure()

    @property
    def selectedCT(self):
        return self._ctComboBox.selectedData

    @property
    def selectedPlanStructure(self):
        return self._planStructureComboBox.selectedData

    def _handlePlanStructure(self, selectedPlanStructure):
        self._objectivesWidget.planDesign = selectedPlanStructure
        if  selectedPlanStructure is not None:
            if isinstance(self._objectivesWidget.planDesign,ProtonPlanDesign):
                self._radiationType = "PROTON"
                self._radiationLabel.setText("PROTON")
                self._radiationLabel.setStyleSheet("background-color: red; color: black;")
                self._beamletGridBox.hide()
                self._spotPlacementBox.show()
            elif isinstance(self._objectivesWidget.planDesign,PhotonPlanDesign):
                self._radiationType = "PHOTON"
                self._radiationLabel.setText("PHOTON")
                self._radiationLabel.setStyleSheet("background-color: yellow; color: black;")
                self._beamletGridBox.show()
                self._spotPlacementBox.hide()
            else: 
                logger.error("Could not identify plan radiation type (planDesign object may be deprecated)")

    def setCurrentPatient(self, patient: Patient):
        self._planStructureComboBox.setPatient(patient)
        self._ctComboBox.setPatient(patient)

        self._objectivesWidget.setPatient(patient)

    def _openConfig(self):
        dialog = OptiSettingsDialog(self._optiConfig)
        if dialog.exec(): dialog.optiParam
        logger.info('opti config = {}'.format(self._optiConfig))
        self._algoBox.setCurrentText(self._optiConfig['method'])

    def _run(self):
        settings = DoseCalculationConfig()
        ctCalibration = readScanner(settings.scannerFolder)

        self.selectedPlanStructure.ct = self.selectedCT
        self.selectedPlanStructure.calibration = ctCalibration

        self._setObjectives()

        # create list of contours
        objROINames = []
        contours = []
        targets = []
        prescription = []
        for obj in self.selectedPlanStructure.objectives.objectivesList:
            # remove duplicate
            if obj.roiName not in objROINames:
                objROINames.append(obj.roiName)
                contours.append(obj.roi)
                if obj.isTarget:
                    targets.append(obj.roi)
                    prescription.append(obj.prescription)

        self.selectedPlanStructure.defineTargetMaskAndPrescription(targets, prescription)

        if self._spotPlacementBox.isChecked():
            self._placeSpots()

        else:
            self._plan = copy.deepcopy(self._plan)
            self._plan.spotMUs = np.ones(self._plan.spotMUs.shape)
            self._plan.planDesign = self.selectedPlanStructure

        self._plan.name = self._planNameEdit.text()

        self._plan.numberOfFractionsPlanned = self._fractionsSpin.value()

        if self._beamletBox.isChecked() and self._beamletBox.isEnabled():
            self._computeBeamlets(contours)

        self._optimize(contours)

    def _setObjectives(self):
        objectiveList = ObjectivesList()
        for obj in self._objectivesWidget.objectives:
            objectiveList.objectivesList.append(obj)

        self.selectedPlanStructure.objectives = objectiveList
        for obj in self.selectedPlanStructure.objectives.objectivesList:
            if obj.robust:
                continue

    def _placeSpots(self):
        self._plan = self.selectedPlanStructure.buildPlan()  # Spot placement

    def _computeBeamlets(self, contours):
        robustOpti = self.selectedPlanStructure.robustness.selectionStrategy != Robustness.Strategies.DISABLED
        if self._radiationType.upper()=="PROTON":
            self._doseCalculationWindow = doseCalculationWindow(self._viewController, self._radiationType, self, contours, beamlets=True, robustOpti=robustOpti)
            self._doseCalculationWindow.setWindowTitle('Beamlet-based configuration')
            self._doseCalculationWindow.setCT(self.selectedPlanStructure.ct)
            self._plan.patient = self.selectedPlanStructure.ct.patient
            self._doseCalculationWindow.setPlan(self._plan)
            self._doseCalculationWindow.exec()
        elif self._radiationType.upper()=="PHOTON":
            self._doseCalculationWindow = doseCalculationWindow(self._viewController, self._radiationType, self, contours, beamlets=True, robustOpti=robustOpti)
            self._doseCalculationWindow.setWindowTitle('Beamlet-based configuration')
            self._doseCalculationWindow.setCT(self.selectedPlanStructure.ct)
            self._plan.patient = self.selectedPlanStructure.ct.patient
            self._doseCalculationWindow.setPlan(self._plan)
            self._doseCalculationWindow.exec()


    def _handleAlgo(self):
        if self._selectedAlgo == "Beamlet-free MCsquare":
            self._beamletBox.setEnabled(False)
            self._configButton.setEnabled(False)
        else:
            self._beamletBox.setEnabled(True)
            if self._selectedAlgo in ["Scipy-LBFGS (recommended)", "Scipy-BFGS", "In-house Gradient", "In-house LBFGS", "In-house BFGS", "FISTA"]:
                self._configButton.setEnabled(True)
                self._optiConfig['method'] = self._selectedAlgo
            else:
                self._configButton.setEnabled(False)

    @property
    def _selectedAlgo(self):
        return self._optiAlgos[self._algoBox.currentIndex()]

    def _optimize(self, contours):
        # Check if robust optimization compatible with solver:
        robustOpti = self.selectedPlanStructure.robustness.selectionStrategy != Robustness.Strategies.DISABLED
        if robustOpti and self._selectedAlgo in ["Beamlet-free MCsquare", "FISTA", "LP"]:
            raise NotImplementedError("Robust optimization is not supported for {} algorithm".format(self._selectedAlgo))

        if self._selectedAlgo == "Beamlet-free MCsquare":
            self._doseCalculationWindow = doseCalculationWindow(self._viewController, self, contours, beamlets=False)
            self._doseCalculationWindow.setWindowTitle('Beamlet-free configuration')
            self._doseCalculationWindow.setCT(self.selectedPlanStructure.ct)
            self._plan.patient = self.selectedPlanStructure.ct.patient
            self._doseCalculationWindow.setPlan(self._plan)
            self._doseCalculationWindow.exec()

        else:
            if self._selectedAlgo == "Scipy-BFGS":
                method = 'Scipy_BFGS'
            if self._selectedAlgo == "Scipy-LBFGS (recommended)":
                method = 'Scipy_L-BFGS-B'
            elif self._selectedAlgo == "In-house BFGS":
                method = 'BFGS'
            elif self._selectedAlgo == "In-house LBFGS":
                method = 'LBFGS'
            elif self._selectedAlgo == "In-house Gradient":
                method = 'Gradient'
            elif self._selectedAlgo == "FISTA":
                method = 'FISTA'
            elif self._selectedAlgo == "LP":
                method = 'LP'

            if self._optiConfig['bounds']:
                solver = BoundConstraintsOptimizer(method = method, plan = self._plan, bounds = self._optiConfig['bounds'], maxiter=self._optiConfig['maxIter'])
            else:
                solver = IntensityModulationOptimizer(method=method, plan=self._plan, maxiter=self._optiConfig['maxIter'])
            # Optimize treatment plan
            doseImage, _ = solver.optimize()
            doseImage.patient = self._doseCalculationWindow._doseComputationPanel.selectedCT.patient
            logger.info("Optimization is done. Check new generated dose image in patient data")

class ObjectivesWidget(QWidget):
    DEFAULT_OBJECTIVES_TEXT = 'No objective defined yet'

    def __init__(self, viewController):
        QWidget.__init__(self)

        self._planDesign = None

        self._roiWindow = ObjectivesWindow(viewController, self)
        self._roiWindow.setMinimumWidth(800)
        self._roiWindow.setMinimumHeight(400)
        self._roiWindow.hide()

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.layout.setContentsMargins(0, 0, 0, 0)

        self._objectivesLabel = QLabel("Objectives:")
        self.layout.addWidget(self._objectivesLabel)

        self._objectivesLabels = QLabel(self.DEFAULT_OBJECTIVES_TEXT)
        self.layout.addWidget(self._objectivesLabels)

        self._objectiveButton = QPushButton('Open objectives panel')
        self._objectiveButton.clicked.connect(self._openObjectivePanel)
        self.layout.addWidget(self._objectiveButton)

        self._roiWindow.objectivesModifiedEvent.connect(self._showObjectives)

    def closeEvent(self, QCloseEvent):
        self._roitTable.objectivesModifiedEvent.disconnect(self._showObjectives)
        super().closeEvent(QCloseEvent)

    @property
    def objectives(self):
        return self._roiWindow.getObjectiveTerms()

    @property
    def planDesign(self):
        return self._planDesign

    @planDesign.setter
    def planDesign(self, pd):
        self._planDesign = pd

        if not (self._planDesign is None):
            self._roiWindow.planDesign = self._planDesign

    def setPatient(self, p: Patient):
        self._roiWindow.patient = p

    def _showObjectives(self):
        objectives = self._roiWindow.getObjectiveTerms()

        if len(objectives) <= 0:
            self._objectivesLabels.setText(self.DEFAULT_OBJECTIVES_TEXT)
            return

        objStr = ''
        for objective in objectives:
            objStr += str(objective.weight)
            objStr += " x "
            objStr += objective.roiName

            if objective.metric == "DMin":
                objStr += ">"
            if objective.metric == "DMax":
                objStr += "<"
            elif objective.metric == "DMaxMean":
                objStr += "="
            objStr += str(objective.limitValue)
            objStr += ' Gy'

            if objective.robust:
                objStr += " (robust)\n"
            else: objStr += '\n'

        self._objectivesLabels.setText(objStr)

    def _openObjectivePanel(self):
        self._roiWindow.planDesign = self.planDesign
        self._roiWindow.show()



import typing

import vtkmodules.vtkRenderingOpenGL2 #This is necessary to avoid a seg fault
import vtkmodules.vtkRenderingFreeType  #This is necessary to avoid a seg fault
from vtkmodules.vtkCommonColor import vtkNamedColors
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import vtkPolyLine, vtkCellArray, vtkPolyData
from vtkmodules.vtkRenderingCore import vtkPolyDataMapper, vtkActor


class CrossHairLayer:
    def __init__(self, renderer, renderWindow):
        self._crossHairActor = vtkActor()
        self._crossHairEnabled = False
        self._crossHairMapper = vtkPolyDataMapper()
        self._position = (0, 0)
        self._renderer = renderer
        self._renderWindow = renderWindow
        self._visible = False

        colors = vtkNamedColors()

        self._crossHairActor.SetMapper(self._crossHairMapper)
        self._crossHairActor.GetProperty().SetColor(colors.GetColor3d('Tomato'))
        self._renderer.AddActor(self._crossHairActor)
        self._crossHairActor.VisibilityOff()

    def close(self):
        pass

    @property
    def position(self) -> typing.Sequence:
        return self._position

    @position.setter
    def position(self, position: typing.Sequence):
        points = vtkPoints()
        points.InsertNextPoint((position[0] - 10, position[1], 0.01))
        points.InsertNextPoint((position[0] + 10, position[1], 0.01))
        points.InsertNextPoint((position[0], position[1] - 10, 0.01))
        points.InsertNextPoint((position[0], position[1] + 10, 0.01))

        polyLine = vtkPolyLine()
        polyLine.GetPointIds().SetNumberOfIds(2)
        polyLine2 = vtkPolyLine()
        polyLine2.GetPointIds().SetNumberOfIds(2)
        for i in range(0, 2):
            polyLine.GetPointIds().SetId(i, i)
            polyLine2.GetPointIds().SetId(i, i + 2)

        cells = vtkCellArray()
        cells.InsertNextCell(polyLine)
        cells.InsertNextCell(polyLine2)

        polyData = vtkPolyData()
        polyData.SetPoints(points)
        polyData.SetLines(cells)

        self._crossHairMapper.SetInputData(polyData)
        self._position = position

        self._renderWindow.Render()

    @property
    def visible(self) -> bool:
        return self._visible

    @visible.setter
    def visible(self, visible: bool):
        self._crossHairActor.SetVisibility(visible)
        self._visible = visible




import numpy as np
from vtkmodules.vtkIOImage import vtkImageImport

from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer


class Image2DForViewer(GenericImageForViewer):

    def __init__(self, image):
        super().__init__(image)

        if hasattr(self, '_dataImporter'):
            return

        # print('in image2DForViewer init', type(image))
        # print('in image2DForViewer init', type(self.data))


        self._dataImporter = vtkImageImport()
        self._selectedPosition = self.data.origin# + self.data.gridSizeInWorldUnit/2.
        self._range = (np.min(self.data.imageArray), np.max(self.data.imageArray))

        self.data.dataChangedSignal.connect(self._updateVTKOutputPort)
        self._updateVTKOutputPort()

    def _updateVTKOutputPort(self):
        print('in image2DForViewer _updateVTKOutputPort')
        print(self.gridSize)
        print(self.spacing)
        print(self.origin)

        shape = self.gridSize  ## dataMultiton magic makes all this available here
        imageOrigin = self.origin
        imageSpacing = self.spacing
        # imageData = np.swapaxes(self.imageArray, 0, 2)
        imageData = self.imageArray
        # imageData[100:140, 200:500] = 300
        imageData -= np.min(imageData)

        # imageData = np.tile(imageData, (3, 1, 1))
        # print(imageData.shape)
        # shape = imageData.shape
        # plt.figure()
        # plt.imshow(imageData)
        # plt.show()
        # shape = imageData.shape
        # print(shape)

        num_array = np.array(np.ravel(imageData), dtype=np.float32)

        self._dataImporter.SetNumberOfScalarComponents(1)
        self._dataImporter.SetDataExtent(0, shape[0] - 1, 0, shape[1] - 1, 0, 0)
        self._dataImporter.SetWholeExtent(0, shape[0] - 1, 0, shape[1] - 1, 0, 0)
        self._dataImporter.SetDataSpacing(imageSpacing[0], imageSpacing[1], 1)
        self._dataImporter.SetDataOrigin(imageOrigin[0], imageOrigin[1], imageOrigin[2])
        self._dataImporter.SetDataScalarTypeToFloat()

        data_string = num_array.tobytes()
        self._dataImporter.CopyImportVoidPointer(data_string, len(data_string))

        self._vtkOutputPort = self._dataImporter.GetOutputPort()

        import vtk

        mapper = vtk.vtkImageMapper()
        print(type(self._vtkOutputPort))
        mapper.SetInputConnection(self._vtkOutputPort)
        # mapper.SetInputData(self._vtkOutputPort)
        actor = vtk.vtkActor2D()
        actor.SetMapper(mapper)

        ren = vtk.vtkRenderer()
        ren.AddActor(actor)
        ren.SetBackground(0.1, 0.2, 0.4)

        renWin = vtk.vtkRenderWindow()
        renWin.AddRenderer(ren)
        renWin.SetSize(400, 400)

        iren = vtk.vtkRenderWindowInteractor()
        iren.SetRenderWindow(renWin)

        renWin.Render()
        iren.Start()

    @property
    def vtkOutputPort(self):
        return self._vtkOutputPort



from PyQt5.QtWidgets import QMainWindow


# TODO A data explorer that shows all kind of data and allow basic operations (rename, save, etc.)
class DataExplorer(QMainWindow):
    def __init__(self, viewController, parent=None):
        super().__init__(parent)

        self._viewController = viewController




from PyQt5.QtWidgets import QWidgetAction, QLabel

from opentps.gui.viewer.dataViewerComponents.dataViewerToolbar import DataViewerToolbar
from opentps.gui.viewer.dataViewerComponents.dvhPlot import DVHViewer


class DVHViewerActions:
    def __init__(self, dvhViewer:DVHViewer):
        self._dvhViewer = dvhViewer

        self._separator = None

        self._doseImageLabel = QLabel('')
        self._doseImageLabel.setFixedHeight(16)
        self._doseImageAction = QWidgetAction(None)
        self._doseImageAction.setDefaultWidget(self._doseImageLabel)

        self._handleDoseImage()

        self.hide()

        self._dvhViewer.doseChangeEvent.connect(self._handleDoseImage)

    def addToToolbar(self, toolbar:DataViewerToolbar):
        self._separator = toolbar.addSeparator()
        toolbar.addAction(self._doseImageAction)

    def hide(self):
        if not self._separator is None:
            self._separator.setVisible(False)
        self._doseImageAction.setVisible(False)

    def show(self):
        self._separator.setVisible(True)
        self._handleDoseImage()

    def _handleDoseImage(self, *args):
        if self._dvhViewer.dose is None:
            self._doseImageAction.setVisible(False)
        else:
            self._doseImageLabel.setText(self._dvhViewer.dose.name)
            #TODO:listen to name change event
            self._doseImageAction.setVisible(True)


from typing import Optional

from vtkmodules.vtkCommonColor import vtkNamedColors
from vtkmodules.vtkRenderingCore import vtkPolyDataMapper, vtkActor

from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer


class PolyData3DLayer_3D:
    def __init__(self, renderer, renderWindow):
        self.imageChangedSignal = Event(object)

        self._image = None
        self._imageToBeSet = None
        self._renderer = renderer
        self._renderWindow = renderWindow

        self._mainActor = vtkActor()
        self._mainMapper = vtkPolyDataMapper()
        self._mainActor.SetMapper(self._mainMapper)

    def close(self):
        self._disconnectAll()
        self._renderer.RemoveActor(self._mainActor)
        self._mainMapper.RemoveAllInputs()

    def update(self):
        self._setImage(self._imageToBeSet)

    @property
    def image(self) -> Optional[Image3DForViewer]:
        """
        Image displayed
        :type:Optional[Image3DForViewer]
        """
        if self._image is None:
            return None

        return self._image

    @image.setter
    def image(self, image:Optional[GenericImageForViewer]):
        if image == self._image:
            return

        if not (isinstance(image, GenericImageForViewer) or (image is None)):
            return

        self._imageToBeSet = image

    def _setImage(self, image:Optional[GenericImageForViewer]):
        self._image = image

        self._disconnectAll()
        self._renderer.RemoveActor(self._mainActor)
        self._mainMapper.RemoveAllInputs()

        if not (self._image is None):
            self._mainMapper.SetInputConnection(self._image.vtkOutputPort)

            self._renderer.AddActor(self._mainActor)

            colors = vtkNamedColors()
            #backFaceProp = vtkProperty()
            #backFaceProp.SetDiffuseColor(colors.GetColor3d("Silver"))
            #self._mainActor.SetBackfaceProperty(backFaceProp)
            self._mainActor.GetProperty().SetDiffuseColor(colors.GetColor3d("NavajoWhite"))

            self._connectAll()

        self.imageChangedSignal.emit(self._image)

        self._renderWindow.Render()


    def _connectAll(self):
        self._image.dataChangedSignal.connect(self._render)

    def _disconnectAll(self):
        if self._image is None:
            return

        self._image.dataChangedSignal.disconnect(self._render)

    def _render(self, *args):
        self._renderWindow.Render()



from PyQt5.QtWidgets import QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QButtonGroup, QRadioButton, QTableWidget, \
    QFileDialog, QPushButton

from opentps.core.io.dataExporter import ExportTypes, ExportConfig, exportPatient


class ExportWindow(QMainWindow):
    def __init__(self, viewController):
        super().__init__()

        self._viewController = viewController

        self.setWindowTitle('Export settings')
        self.resize(400, 400)

        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        self._layout = QVBoxLayout()
        centralWidget.setLayout(self._layout)

        self._exportTable = ExportTable(parent=self)
        self._layout.addWidget(self._exportTable)

        self._exportButton = QPushButton("Select folder and export")
        self._exportButton.clicked.connect(self._handleExport)
        self._layout.addWidget(self._exportButton)

        self._layout.addStretch()
        self.adjustSize()
        self.setFixedSize(self.size())

    def _handleExport(self):
        # TODO: Use export options defined in ExportTable

        folderpath = QFileDialog.getExistingDirectory(self, 'Select folder')

        if folderpath == "":
            return

        exportPatient(self._viewController.currentPatient, folderpath, self._exportTable.exportConfig)

class ExportTable(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        dataTypes = ExportConfig()
        self._exportTypes = [ExportTypes.DICOM, ExportTypes.MHD, ExportTypes.MCSQUARE, ExportTypes.PICKLE]
        self._buttonGroups = []
        self._radio_buttons = {}  # To keep track of the state of radio buttons

        rowNb = len(dataTypes)
        colNb = len(self._exportTypes)

        self.table = QTableWidget()
        self.table.setRowCount(rowNb)
        self.table.setColumnCount(colNb)
        self.table.setHorizontalHeaderLabels([exportType.value for exportType in self._exportTypes])
        self.table.setVerticalHeaderLabels([dataType.name for dataType in dataTypes])

        self.table.setFixedWidth((colNb+1)*self.table.columnWidth(0))

        self._layout = QHBoxLayout(self)
        self._layout.addWidget(self.table)

        for row, dataType in enumerate(dataTypes):
            button_group = QButtonGroup(self)
            button_group.setExclusive(False) 
            self._buttonGroups.append(button_group)

            for col, exportType in enumerate(self._exportTypes):
                checkbox = QRadioButton()
                button_group.addButton(checkbox, col)
                self.table.setCellWidget(row, col, checkbox)

                if exportType in dataType.exportTypes:
                    checkbox.setEnabled(True)
                    checkbox.setChecked(False) 
                    self._radio_buttons[(row, col)] = checkbox  
                    checkbox.toggled.connect(lambda checked, cb=checkbox, bg=button_group: self._toggle_radio_button(cb, bg, checked))
                else:
                    checkbox.setChecked(False)
                    checkbox.setEnabled(False)

    def _toggle_radio_button(self, checkbox, button_group, checked):
        if not checked:
            return  

        for btn in button_group.buttons():
            if btn != checkbox:
                btn.setChecked(False)

    @property
    def exportConfig(self) -> ExportConfig:
        config = ExportConfig()

        for i, dataType in enumerate(config):
            checked_id = self._buttonGroups[i].checkedId()
            if checked_id != -1:
                dataType.exportType = self._exportTypes[checked_id]
            else:
                dataType.exportType = None

        return config


from typing import Sequence, Tuple

import numpy as np
import vtkmodules.vtkCommonCore as vtkCommonCore
from matplotlib import pyplot as plt
from vtkmodules.vtkCommonDataModel import vtkPiecewiseFunction
from vtkmodules.vtkRenderingCore import vtkColorTransferFunction


def uniqueColorLT(threshold:float, opacity:float, color:Sequence[float]) -> vtkCommonCore.vtkLookupTable:
    table = vtkCommonCore.vtkLookupTable()
    table.SetRange(threshold-1., threshold)  # image intensity range
    table.SetValueRange(0.0, 1.0)  # from black to white
    table.SetSaturationRange(0.0, 0.0)  # no color saturation
    table.SetRampToLinear()

    table.SetNumberOfTableValues(2)
    table.SetTableValue(0, (0, 0, 0, 0))
    table.SetTableValue(1, (color[0], color[1], color[2], opacity))

    table.SetBelowRangeColor(0, 0, 0, 0)
    table.SetUseBelowRangeColor(True)
    table.SetAboveRangeColor(color[0], color[1], color[2], opacity)
    table.SetUseAboveRangeColor(True)
    table.Build()

    return table

def uniqueColorLTTo3DLT(lt:vtkCommonCore.vtkLookupTable) -> Tuple[vtkColorTransferFunction, vtkPiecewiseFunction, vtkPiecewiseFunction]:
    rangeVal = lt.GetRange()
    opacity = lt.GetOpacity(rangeVal[1])

    volumeColor = vtkColorTransferFunction()
    volumeScalarOpacity = vtkPiecewiseFunction()
    volumeGradientOpacity = vtkPiecewiseFunction()

    volumeScalarOpacity.AddPoint(rangeVal[0], 0.)
    volumeScalarOpacity.AddPoint(rangeVal[1], opacity)

    volumeGradientOpacity.AddPoint(rangeVal[0], 0.)
    volumeGradientOpacity.AddPoint(rangeVal[1], opacity)

    tableVals = np.linspace(rangeVal[0], rangeVal[1], lt.GetNumberOfTableValues())
    for i in range(lt.GetNumberOfTableValues()):
        tbVal = lt.GetTableValue(i)
        volumeColor.AddRGBPoint(tableVals[i], tbVal[0], tbVal[1], tbVal[2])

    return volumeColor, volumeScalarOpacity, volumeGradientOpacity


def fusionLT(range:Sequence[float], opacity:float, colormap:str) -> vtkCommonCore.vtkLookupTable:
    table = vtkCommonCore.vtkLookupTable()
    table.SetRange(range[0], range[1])  # image intensity range
    table.SetValueRange(0.0, 1.0)  # from black to white
    table.SetSaturationRange(0.0, 0.0)  # no color saturation
    table.SetRampToLinear()
    cm = plt.get_cmap(colormap)
    linInd = list(np.arange(0, 1.01, 0.01))

    table.SetNumberOfTableValues(len(linInd))
    LastCMVal = (0, 0, 0)
    for i, ind in enumerate(linInd):
        cmVal = cm(ind)
        LastCMVal = cmVal
        if i==0:
            table.SetTableValue(i, (cmVal[0], cmVal[1], cmVal[2], 0))
        else:
            table.SetTableValue(i, (cmVal[0], cmVal[1], cmVal[2], opacity))

    table.SetBelowRangeColor(0, 0, 0, 0)
    table.SetUseBelowRangeColor(True)
    table.SetAboveRangeColor(LastCMVal[0], LastCMVal[1], LastCMVal[2], opacity)
    table.SetUseAboveRangeColor(True)
    table.Build()

    return table

def grayLT(range) -> vtkCommonCore.vtkLookupTable:
    table = vtkCommonCore.vtkLookupTable()
    table.SetRange(range[0], range[1])  # image intensity range
    table.SetTableRange(range[0], range[1])  # image intensity range
    table.SetValueRange(0.0, 1.0)  # from black to white
    table.SetSaturationRange(0.0, 0.0)  # no color saturation
    table.SetRampToLinear()
    table.SetAlpha(1.)
    table.SetAboveRangeColor(1., 1., 1., 1.)
    table.SetBelowRangeColor(0., 0., 0., 1.)
    table.SetUseAboveRangeColor(True)
    table.SetUseBelowRangeColor(True)
    table.Build()

    return table


def fusionLTTo3DLT(lt:vtkCommonCore.vtkLookupTable) -> Tuple[vtkColorTransferFunction, vtkPiecewiseFunction, vtkPiecewiseFunction]:
    rangeVal = lt.GetRange()
    opacity0 = lt.GetOpacity(0)
    opacity1 = lt.GetOpacity(1)

    volumeColor = vtkColorTransferFunction()
    volumeScalarOpacity = vtkPiecewiseFunction()
    volumeGradientOpacity = vtkPiecewiseFunction()

    volumeScalarOpacity.AddPoint(rangeVal[0], opacity0)
    volumeScalarOpacity.AddPoint((rangeVal[0]+rangeVal[1])/.2, opacity0)
    volumeScalarOpacity.AddPoint(3*(rangeVal[0] + rangeVal[1]) /4. , opacity1/2.)
    volumeScalarOpacity.AddPoint(rangeVal[1], opacity1)

    volumeGradientOpacity.AddPoint(rangeVal[0], opacity1)
    volumeGradientOpacity.AddPoint((rangeVal[0]+rangeVal[1])/2, opacity1/4.)
    volumeGradientOpacity.AddPoint(3*(rangeVal[0] + rangeVal[1]) /4. , opacity1/2.)
    volumeGradientOpacity.AddPoint(rangeVal[1], opacity1)

    tableVals = np.linspace(rangeVal[0], rangeVal[1], lt.GetNumberOfTableValues())
    for i in range(lt.GetNumberOfTableValues()):
        tbVal = lt.GetTableValue(i)
        volumeColor.AddRGBPoint(tableVals[i], tbVal[0], tbVal[1], tbVal[2])

    return volumeColor, volumeScalarOpacity, volumeGradientOpacity

def ct3DLT() -> Tuple[vtkColorTransferFunction, vtkPiecewiseFunction, vtkPiecewiseFunction]:
    volumeColor = vtkColorTransferFunction()
    volumeScalarOpacity = vtkPiecewiseFunction()
    volumeGradientOpacity = vtkPiecewiseFunction()

    volumeColor.AddRGBPoint(-1000, 0.0, 0.0, 0.0)
    volumeColor.AddRGBPoint(-500, 240.0/255.0, 184.0/255.0, 160.0/255.0)
    volumeColor.AddRGBPoint(0, 232/255, 147/255, 132/255)
    volumeColor.AddRGBPoint(500, 1.0, 1.0, 240.0 / 255.0)
    volumeColor.AddRGBPoint(700,  242/255, 220/255, 172/255)
    volumeColor.AddRGBPoint(1800, 173/255, 166/255, 166/255)
    volumeColor.AddRGBPoint(2500, 237/255, 171/255, 64/255)

    volumeScalarOpacity.AddPoint(-1000, 0.00)
    volumeScalarOpacity.AddPoint(-500, 0.05)
    volumeScalarOpacity.AddPoint(-200, 0.5)
    volumeScalarOpacity.AddPoint(-100, 0.85)
    volumeScalarOpacity.AddPoint(0, 0.85)
    volumeScalarOpacity.AddPoint(500, 0.85)
    volumeScalarOpacity.AddPoint(700, 0.85)

    volumeGradientOpacity.AddPoint(-1000, 0.0)
    volumeGradientOpacity.AddPoint(-500, 0.0)
    volumeGradientOpacity.AddPoint(0, 0.6)
    volumeGradientOpacity.AddPoint(500, 0.85)

    return volumeColor, volumeScalarOpacity, volumeGradientOpacity


from typing import Optional, Union

from PyQt5.QtWidgets import *

import vtkmodules.vtkRenderingCore as vtkRenderingCore
from vtkmodules import vtkInteractionStyle
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor

from opentps.core.data import ROIContour
from opentps.core.data.images import ROIMask
from opentps.core.data.images._image3D import Image3D
from opentps.core.data.plan import RTPlan
from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.contourLayer_3D import ContourLayer_3D
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.primaryImage3DLayer_3D import PrimaryImage3DLayer_3D
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.rtplanLayer_3D import RTPlanLayer_3D
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.secondaryImage3DLayer_3D import SecondaryImage3DLayer_3D
from opentps.gui.viewer.dataViewerComponents.blackEmptyPlot import BlackEmptyPlot


class Image3DViewer_3D(QWidget):
    def __init__(self, viewController, parent=None):
        QWidget.__init__(self, parent=parent)

        self.primaryImageSignal = Event(object)
        self.secondaryImageSignal = Event(object)

        self._blackWidget = BlackEmptyPlot(self)
        self._renderer = vtkRenderingCore.vtkRenderer()
        self._viewController = viewController
        self._vtkWidget = QVTKRenderWindowInteractor(self)
        self._mainLayout = QVBoxLayout()
        self._renderWindow = self._vtkWidget.GetRenderWindow()

        self._iStyle = vtkInteractionStyle.vtkInteractorStyleTrackballCamera()

        self._primaryImageLayer = PrimaryImage3DLayer_3D(self._renderer, self._renderWindow, self._iStyle)
        self._secondaryImageLayer = SecondaryImage3DLayer_3D(self._renderer, self._renderWindow, self._iStyle)
        self._contourLayer = ContourLayer_3D(self._renderer, self._renderWindow, self._iStyle)
        self._rtPlanLayer = RTPlanLayer_3D(self._renderer, self._renderWindow)


        self.setLayout(self._mainLayout)
        self._vtkWidget.hide()
        self._mainLayout.addWidget(self._blackWidget)
        self._blackWidget.show()
        self._mainLayout.setContentsMargins(0, 0, 0, 0)

        self._renderer.SetBackground(0, 0, 0)
        self._renderer.GetActiveCamera().SetParallelProjection(True)

        self._renderWindow.GetInteractor().SetInteractorStyle(self._iStyle)
        self._renderWindow.AddRenderer(self._renderer)

        self._closed = False

    def closeEvent(self, QCloseEvent):
        if self._closed:
            return

        self._closed = True
        self.reset()
        self._renderWindow.Finalize()
        self._vtkWidget.close()
        del self._renderWindow, self._vtkWidget
        super().closeEvent(QCloseEvent)

    def reset(self):
        self._rtPlanLayer.close()
        self._contourLayer.close()
        self._primaryImageLayer.close()
        self._secondaryImageLayer.close()

    def show(self):
        super(Image3DViewer_3D, self).show()
        self.update()
        self._primaryImageLayer.update()
        self._secondaryImageLayer.update()
        self._rtPlanLayer.update()
        self._contourLayer.update()

    def update(self):
        self._primaryImageLayer.update()
        self._secondaryImageLayer.update()
        self._rtPlanLayer.update()


    @property
    def primaryImage(self) -> Optional[Image3D]:
        if self._primaryImageLayer.image is None:
            return None
        return self._primaryImageLayer.image.data

    @primaryImage.setter
    def primaryImage(self, image: Image3D):
        self._setPrimaryImageForViewer(Image3DForViewer(image))
        if self.isVisible():
            self.update()

        self.primaryImageSignal.emit(self.primaryImage)

    def _resetPrimaryImageLayer(self):
        self._primaryImageLayer.image = None
        self._mainLayout.removeWidget(self._vtkWidget)
        self._vtkWidget.hide()
        self._mainLayout.addWidget(self._blackWidget)
        self._blackWidget.show()

    def _setPrimaryImageForViewer(self, image:GenericImageForViewer):
        self._primaryImageLayer.image = image

        self._blackWidget.hide()
        self._mainLayout.removeWidget(self._blackWidget)
        self._mainLayout.addWidget(self._vtkWidget)
        self._vtkWidget.show()

    @property
    def secondaryImage(self) -> Optional[Image3D]:
        if self._secondaryImageLayer.image is None:
            return None
        return self._secondaryImageLayer.image.data

    @secondaryImage.setter
    def secondaryImage(self, image: Image3D):
        if self.primaryImage is None:
            return

        if image is None:
            self._secondaryImageLayer.image = None
            if self.isVisible():
                self.update()

            self.secondaryImageSignal.emit(self.secondaryImage)
            return
        else:
            self._secondaryImageLayer.image = Image3DForViewer(image)

        if self.isVisible():
            self.update()

        self._renderWindow.Render()

        self.secondaryImageSignal.emit(self.secondaryImage)

    @property
    def rtPlan(self) -> RTPlan:
        raise NotImplementedError

    @rtPlan.setter
    def rtPlan(self, plan: RTPlan):
        self._rtPlanLayer.setPlan(plan)

        if self.isVisible():
            self.update()

    def setNewContour(self, contour:Union[ROIContour, ROIMask]):
        self._contourLayer.setNewContour(contour)

        if self.isVisible():
            self._contourLayer.update()




import os
from PyQt5.QtWidgets import QMainWindow, QHBoxLayout, QWidget
from PyQt5.QtGui import QIcon

from opentps.gui.panels.mainToolbar import MainToolbar
from opentps.gui.viewer.viewerPanel import ViewerPanel
from opentps.gui.programSettingEditor import ProgramSettingEditor
from opentps.gui.statusBar import StatusBar
from opentps.core.utils.programSettings import ProgramSettings
import opentps.gui.res.icons as IconModule

class MainWindow(QMainWindow):
    def __init__(self, viewControler):
        QMainWindow.__init__(self)

        self.setWindowTitle('OpenTPS')
        self.setWindowIcon(QIcon(IconModule.__path__[0] + os.path.sep + 'OpenTPS_icon.png'))
        self.resize(1400, 920)

        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        self.mainLayout = QHBoxLayout()  ## not sure the "self" is necessary for mainLayout, it shoudnt be called outside this constructor
        centralWidget.setLayout(self.mainLayout)

        self._viewControler = viewControler

        # create and add the tool panel on the left
        self.toolbox_width = 270
        self.mainToolbar = MainToolbar(self._viewControler)
        self.mainToolbar.setFixedWidth(self.toolbox_width)
        self.mainLayout.addWidget(self.mainToolbar)

        ProgramSettingEditor.setProgramSettings(ProgramSettings())
        ProgramSettingEditor.setMainToolbar(self.mainToolbar)

        # create and add the viewer panel
        self.viewerPanel = ViewerPanel(self._viewControler, self)
        self.mainLayout.addWidget(self.viewerPanel)

        self.statusBar = StatusBar()
        self.setStatusBar(self.statusBar)
        self.statusBar.show()

    @property
    def viewController(self):
        return self._viewControler

    def closeEvent(self, QCloseEvent):
        self.viewerPanel.close()
        super().closeEvent(QCloseEvent)


    def setLateralToolbar(self, toolbar):
        self.mainLayout.addWidget(toolbar)
        toolbar.setFixedWidth(self.toolbox_width)

    def setMainPanel(self, mainPanel):
        self.mainLayout.addWidget(mainPanel)



import numpy as np
from opentps.gui.viewer.dataForViewer.vtkSimpleImageFilter import VtkSimpleImageFilter

from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image2DForViewer import Image2DForViewer


class Dyn2DSeqForViewer(GenericImageForViewer):
    def __init__(self, dyn2DSeq):
        super().__init__(dyn2DSeq)

        if hasattr(self, '_image3DForViewerList'):
            return

        dyn2DSeq = self.data
        self._selectedPosition = np.array(dyn2DSeq.dyn2DImageList[0].origin) + np.array(dyn2DSeq.dyn2DImageList[0].gridSize) * np.array(dyn2DSeq.dyn2DImageList[0].spacing) / 2.0

        # This creates all image3DForViewers within the image sequence with the side effect that all VTK output port are initialized.
        self._image2DForViewerList = self._getImg2DForViewerList(dyn2DSeq.dyn2DImageList)

        self._simpleFilter = VtkSimpleImageFilter()
        self._currentIndexIn2DSeq = 0
        self._updateVTKOutputPort()
        self._vtkOutputPort = self._simpleFilter.GetOutputPort()
        self._range = (np.min(self.data.dyn2DImageList[0].imageArray), np.max(self.data.dyn2DImageList[0].imageArray))

    def _getImg2DForViewerList(self, dyn2DSeqImgList):
        vtkImageList = []
        for image in dyn2DSeqImgList:
            vtkImageList.append(Image2DForViewer(image))

        return vtkImageList

    @property
    def currentIndexIn2DSeq(self):
        return self._currentIndexIn2DSeq

    @currentIndexIn2DSeq.setter
    def currentIndexIn2DSeq(self, ind):
        if self._currentIndexIn2DSeq == ind:
            return

        self._currentIndexIn2DSeq = ind
        self._updateVTKOutputPort()

    def _updateVTKOutputPort(self):
        self._simpleFilter.RemoveAllInputs()
        currImageForViewer = self._image2DForViewerList[self._currentIndexIn2DSeq]
        self._simpleFilter.SetInputConnection(currImageForViewer.vtkOutputPort)

    @property
    def vtkOutputPort(self):
        return self._vtkOutputPort




import typing
from math import sqrt

import numpy as np
from pyqtgraph import mkPen
from vtkmodules.vtkInteractionWidgets import vtkLineWidget2

from opentps.core import Event
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.contourLayer import ContourLayer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.primaryImage3DLayer import PrimaryImage3DLayer


class ProfileWidget:
    class ProfileWidgetCallback:
        def __init__(self):
            self._setPrimaryImageData = lambda *args, **kwargs: None
            self._setSecondaryImageData = lambda *args, **kwargs: None
            self._setContourData = lambda *args, **kwargs: None

        @property
        def setPrimaryImageData(self):
            return self._setPrimaryImageData

        @setPrimaryImageData.setter
        def setPrimaryImageData(self, func):
            self._setPrimaryImageData = func

        @property
        def setSecondaryImageData(self):
            return self._setSecondaryImageData

        @setSecondaryImageData.setter
        def setSecondaryImageData(self, func):
            self._setSecondaryImageData = func

        @property
        def setContourData(self):
            return self._setContourData

        @setContourData.setter
        def setContourData(self, func):
            self._setContourData = func

    def __init__(self, renderer, renderWindow):
        self.lineWidgeEnabledSignal = Event(bool)

        self._lineWidget = vtkLineWidget2()
        self._lineWidgetCallback = None
        self._lineWidgetEnabled = False
        self._primaryLayer = None
        self._secondaryLayer = None
        self._contourLayer = None
        self._renderer = renderer
        self._renderWindow = renderWindow

        self._lineWidget.SetCurrentRenderer(self._renderer)
        self._lineWidget.AddObserver("InteractionEvent", self.onProfileWidgetInteraction)
        self._lineWidget.AddObserver("EndInteractionEvent", self.onProfileWidgetInteraction)
        self._lineWidget.SetInteractor(self._renderWindow.GetInteractor())

    @property
    def enabled(self) -> bool:
        return self._lineWidgetEnabled

    @enabled.setter
    def enabled(self, enabled: bool):
        if enabled == self._lineWidgetEnabled:
            return

        if enabled:
            self._lineWidget.On()
            self._lineWidget.GetLineRepresentation().SetLineColor(1, 0, 0)
            self._lineWidgetEnabled = True
        else:
            self._lineWidget.Off()
            self._lineWidgetEnabled = False
            self._renderWindow.Render()

        self.lineWidgeEnabledSignal.emit(self._lineWidgetEnabled)

    @property
    def callback(self):
        return self._lineWidgetCallback

    @callback.setter
    def callback(self, method):
        self._lineWidgetCallback = method

    @property
    def primaryLayer(self):
        return self._primaryLayer

    @primaryLayer.setter
    def primaryLayer(self, layer):
        self._primaryLayer = layer

        if isinstance(self._primaryLayer, PrimaryImage3DLayer):
            if not self._primaryLayer.image is None:
                self._primaryLayer._reslice.RemoveObserver(self._endEventObserver)

            self._endEventObserver = self._primaryLayer._reslice.AddObserver("EndEvent", self.onProfileWidgetInteraction)

    @property
    def secondaryLayer(self):
        return self._secondaryLayer

    @secondaryLayer.setter
    def secondaryLayer(self, layer):
        self._secondaryLayer = layer

    @property
    def contourLayer(self):
        return self._secondaryLayer

    @contourLayer.setter
    def contourLayer(self, layer):
        self._contourLayer = layer

    def setInitialPosition(self, worldPos: typing.Sequence):
        self._lineWidget.GetLineRepresentation().SetPoint1WorldPosition((worldPos[0], worldPos[1], 0.01))
        self._lineWidget.GetLineRepresentation().SetPoint2WorldPosition((worldPos[0], worldPos[1], 0.01))

    def onProfileWidgetInteraction(self, obj, event):
        if not self.enabled:
            return

        point1 = self._lineWidget.GetLineRepresentation().GetPoint1WorldPosition()
        point2 = self._lineWidget.GetLineRepresentation().GetPoint2WorldPosition()

        if point1[1]==point1[2]==point2[1]==point2[2]:
            return

        matrix = self._primaryLayer.resliceAxes
        point1 = matrix.MultiplyPoint((point1[0], point1[1], 0, 1))
        point2 = matrix.MultiplyPoint((point2[0], point2[1], 0, 1))

        x, y = self._resliceLayerDataBewteenTwoPoints(self._primaryLayer, point1, point2)
        self._lineWidgetCallback.setPrimaryImageData(x, y, name=self._layerImageName(self._primaryLayer))

        x, y = self._resliceLayerDataBewteenTwoPoints(self._secondaryLayer, point1, point2)
        self._lineWidgetCallback.setSecondaryImageData(x, y, name=self._layerImageName(self._secondaryLayer))

        contourNames = [contour.name for contour in self._contourLayer.contours]
        contourData = self._resliceContoursBetweenTwoPoints(self._contourLayer, point1, point2)
        pen = [mkPen(color=contour.color, width=1) for contour in self._contourLayer.contours]
        self._lineWidgetCallback.setContourData(contourData, name=contourNames, pen=pen)

    def _resliceLayerDataBewteenTwoPoints(self, layer, point1, point2):
        if layer.image is None:
            return ([0, 0], [0, 0])

        num = 1000
        points0 = np.linspace(point1[0], point2[0], num)
        points1 = np.linspace(point1[1], point2[1], num)
        points2 = np.linspace(point1[2], point2[2], num)
        data = np.array(points1)

        x = np.linspace(0, sqrt((point2[0] - point1[0]) * (point2[0] - point1[0]) + (point2[1] - point1[1]) * (
                point2[1] - point1[1]) + (point2[2] - point1[2]) * (point2[2] - point1[2])), num)

        for i, p0 in enumerate(points0):
            data[i] = layer.resliceDataFromPhysicalPoint((p0, points1[i], points2[i]))

        return (x, data)

    def _layerImageName(self, layer):
        if layer.image is None:
            return None

        return layer.image.name


    def _resliceContoursBetweenTwoPoints(self, layer:ContourLayer, point1:typing.Sequence[float], point2:typing.Sequence[float]):
        contours = layer.contours

        res = []
        if len(contours)<=0:
            for contour in contours:
                res.append(([0, 0], [0, 0]))
            return res

        num = 1000
        points0 = np.linspace(point1[0], point2[0], num)
        points1 = np.linspace(point1[1], point2[1], num)
        points2 = np.linspace(point1[2], point2[2], num)


        data = np.zeros((len(contours), num))
        x = np.linspace(0, sqrt((point2[0] - point1[0]) * (point2[0] - point1[0]) + (point2[1] - point1[1]) * (
                point2[1] - point1[1]) + (point2[2] - point1[2]) * (point2[2] - point1[2])), num)


        for i, p0 in enumerate(points0):
            contoursData = layer.resliceDataFromPhysicalPoint((p0, points1[i], points2[i]))
            for c, contour in enumerate(contours):
                data[c, i] = -10000+contoursData[c]*20000

        for c, contour in enumerate(contours):
            res.append((x, data[c, :]))

        return res



import functools
from typing import Sequence

import matplotlib
from PyQt5.QtWidgets import QComboBox, QWidgetAction, QMenu, QAction

from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer
from opentps.gui.viewer.dataViewerComponents.dataViewerToolbar import DataViewerToolbar
from opentps.gui.viewer.dataViewerComponents.doseComparisonImageProvider import DoseComparisonImageProvider
from opentps.gui.viewer.dataViewerComponents.image3DViewer import Image3DViewer
from opentps.gui.viewer.dataViewerComponents.imageFusionPropEditor import ImageFusionPropEditor


class ImageViewerActions:
    def __init__(self, imageViewer:Image3DViewer):
        self._imageViewer = imageViewer

        self._viewTypeToStr = {self._imageViewer.ViewerTypes.AXIAL: 'Axial',
                               self._imageViewer.ViewerTypes.CORONAL: 'Coronal',
                               self._imageViewer.ViewerTypes.SAGITTAL: 'Sagittal'}
        self._strToViewType = {v: k for k, v in self._viewTypeToStr.items()}

        self._separator = None

        self._viewTypeCombo = QComboBox()
        self._viewTypeCombo.setFixedSize(80, 16)
        self._viewTypeCombo.addItems(list(self._viewTypeToStr.values()))

        self._viewTypeAction = QWidgetAction(None)
        self._viewTypeAction.setDefaultWidget(self._viewTypeCombo)

        self._primaryImageMenu = PrimaryImageMenu(self._imageViewer)
        self._secondaryImageMenu = SecondaryImageMenu(self._imageViewer)
        self._doseComparisonMenu = DoseComparisonMenu(self._imageViewer)
        self._rtPlanMenu = RTPlanMenu(self._imageViewer)

        self.hide()

        self._viewTypeCombo.setCurrentIndex(self._viewTypeToIndex(self._imageViewer.viewType))
        self._viewTypeCombo.currentIndexChanged.connect(self._handleViewTypeSelection)
        self._imageViewer.viewTypeChangedSignal.connect(self._handleExternalViewTypeChange)

    @property
    def doseComparisonDataViewer(self):
        return self._doseComparisonMenu.dataViewer

    @doseComparisonDataViewer.setter
    def doseComparisonDataViewer(self, dataViewer):
        self._doseComparisonMenu.dataViewer = dataViewer


    def _viewTypeToIndex(self, viewType):
        return list(self._viewTypeToStr.keys()).index(viewType)

    def setImageViewer(self, imageViewer):
        self._imageViewer = imageViewer

    def addToToolbar(self, toolbar:DataViewerToolbar):
        self._separator = toolbar.addSeparator()
        toolbar.addAction(self._viewTypeAction)
        toolbar.toolsMenu.addMenu(self._primaryImageMenu)
        toolbar.toolsMenu.addMenu(self._secondaryImageMenu)
        toolbar.toolsMenu.addMenu(self._doseComparisonMenu)
        toolbar.toolsMenu.addMenu(self._rtPlanMenu)

    def hide(self):
        if not self._separator is None:
            self._separator.setVisible(False)
        self._viewTypeAction.setVisible(False)

    def show(self):
        self._separator.setVisible(True)
        self._viewTypeAction.setVisible(True)

    def _handleViewTypeSelection(self, selectionIndex):
        selectionText = self._viewTypeCombo.itemText(selectionIndex)
        self._imageViewer.viewType = self._strToViewType[selectionText]

    def _handleExternalViewTypeChange(self, viewType):
        self._viewTypeCombo.setCurrentIndex(self._viewTypeToIndex(viewType))


class PrimaryImageMenu(QMenu):
    def __init__(self, imageViewer:Image3DViewer):
        super().__init__("Primary image")

        self._imageViewer = imageViewer

        self._wwlMenu = QMenu("Window level/width", self)
        self.addMenu(self._wwlMenu)

        self._presetMenu = QMenu("Presets", self)
        self._wwlMenu.addMenu(self._presetMenu)

        self._wwlActions = []
        for ps in presets():
            wwlAction = QAction(ps.name + ' ' + str((ps.wwl[1], ps.wwl[0])), self)
            wwlAction.triggered.connect(functools.partial(self._setPreset, ps))

            self._wwlActions.append(wwlAction)

            self._presetMenu.addAction(wwlAction)

    def _setPreset(self, ps):
        Image3DForViewer(self._imageViewer.primaryImage).wwlValue = ps.wwl

class SecondaryImageMenu(QMenu):
    def __init__(self, imageViewer:Image3DViewer):
        super().__init__("Secondary image")

        self._imageViewer = imageViewer
        self._secondaryImageLayer = self._imageViewer.secondaryImageLayer

        self._resetAction = QAction("Reset", self)
        self._resetAction.triggered.connect(self._resetImage)
        self.addAction(self._resetAction)

        self._colorMapMenu = QMenu("Colormap", self)
        self.addMenu(self._colorMapMenu)

        self._colormapActions = []
        cms = matplotlib.pyplot.colormaps()

        for cm in cms:
            cmAction = QAction(cm, self._colorMapMenu)
            cmAction.triggered.connect(functools.partial(self.setFusion, cm))
            self._colorMapMenu.addAction(cmAction)
            self._colormapActions.append(cmAction)

        self._colorbarAction = QAction("Show/hide colorbar", self)
        self._colorbarAction.triggered.connect(self._setColorbarOnOff)
        self.addAction(self._colorbarAction)

        self._wwlAction = QAction("Window level/width", self)
        self._wwlAction.triggered.connect(self._showImageProperties)
        self.addAction(self._wwlAction)

    def _resetImage(self):
        self._imageViewer.secondaryImage = None

    def _setColorbarOnOff(self):
        self._secondaryImageLayer.colorbarOn = not self._secondaryImageLayer.colorbarOn

    def _showImageProperties(self):
        self._imageFusionProp = ImageFusionPropEditor(self._secondaryImageLayer.image.data)
        self._imageFusionProp.show()

    def setFusion(self, name:str):
        self._imageViewer.secondaryImageLayer.image.lookupTableName = name

class DoseComparisonMenu(QMenu):
    def __init__(self, imageViewer:Image3DViewer):
        super().__init__("Dose comparison")

        self.dataViewer = None

        self._imageViewer = imageViewer
        self._secondaryImageLayer = self._imageViewer.secondaryImageLayer

        self._metricsMenu = QMenu("Metrics", self)
        self.addMenu(self._metricsMenu)

        self._diffAction = QAction("Difference", self._metricsMenu)
        self._diffAction.triggered.connect(self._setDiffMetric)
        self._metricsMenu.addAction(self._diffAction)
        self._absDiffAction = QAction("Absolute difference", self._metricsMenu)
        self._absDiffAction.triggered.connect(self._setAbsDiffMetric)
        self._metricsMenu.addAction(self._absDiffAction)
        self._gammaAction = QAction("Gamma", self._metricsMenu)
        self._gammaAction.triggered.connect(self._setGammaMetric)
        self._metricsMenu.addAction(self._gammaAction)

        self._colorMapMenu = QMenu("Colormap", self)
        self.addMenu(self._colorMapMenu)

        self._colormapActions = []
        cms = matplotlib.pyplot.colormaps()

        for cm in cms:
            cmAction = QAction(cm, self._colorMapMenu)
            cmAction.triggered.connect(functools.partial(self.setFusion, cm))
            self._colorMapMenu.addAction(cmAction)
            self._colormapActions.append(cmAction)

        self._colorbarAction = QAction("Show/hide colorbar", self)
        self._colorbarAction.triggered.connect(self._setColorbarOnOff)
        self.addAction(self._colorbarAction)

        self._wwlAction = QAction("Window level/width", self)
        self._wwlAction.triggered.connect(self._showImageProperties)
        self.addAction(self._wwlAction)

    def _setColorbarOnOff(self):
        self._secondaryImageLayer.colorbarOn = not self._secondaryImageLayer.colorbarOn

    def _showImageProperties(self):
        self._imageFusionProp = ImageFusionPropEditor(self._secondaryImageLayer.image.data)
        self._imageFusionProp.show()

    def _setDiffMetric(self):
        self.dataViewer.comparisonMetric = DoseComparisonImageProvider.Metric.DIFFERENCE
    def _setAbsDiffMetric(self):
        self.dataViewer.comparisonMetric = DoseComparisonImageProvider.Metric.ABSOLUTE_DIFFERENCE
    def _setGammaMetric(self):
        self.dataViewer.comparisonMetric = DoseComparisonImageProvider.Metric.GAMMA

    def setFusion(self, name:str):
        self._imageViewer.secondaryImageLayer.lookupTableName = name

class RTPlanMenu(QMenu):
    def __init__(self, imageViewer:Image3DViewer):
        super().__init__("RT plan")

        self._imageViewer = imageViewer

        self._resetAction = QAction("Reset", self)
        self._resetAction.triggered.connect(self._resetPlan)
        self.addAction(self._resetAction)

    def _resetPlan(self):
        self._imageViewer.rtPlan = None

class WWLPreset:
    def __init__(self, name:str, wwl:Sequence[float]):
        self.name:str = name
        self.wwl:Sequence[float] = wwl

def presets() -> Sequence[WWLPreset]:
    presets = []

    ps = WWLPreset("Bone", (1600, 450))
    presets.append(ps)
    ps = WWLPreset("Brain", (100, 35))
    presets.append(ps)
    ps = WWLPreset("Dental", (2000, 400))
    presets.append(ps)
    ps = WWLPreset("Inner ear", (4000, 700))
    presets.append(ps)
    ps = WWLPreset("Larynx", (250, 40))
    presets.append(ps)
    ps = WWLPreset("Liver", (350, 50))
    presets.append(ps)
    ps = WWLPreset("Lung", (1600, -600))
    presets.append(ps)
    ps = WWLPreset("Mediastinum", (400, 40))
    presets.append(ps)
    ps = WWLPreset("Pelvis", (1000, 250))
    presets.append(ps)
    ps = WWLPreset("Soft tissue", (350, 40))
    presets.append(ps)
    ps = WWLPreset("Spine", (300, 35))
    presets.append(ps)
    ps = WWLPreset("Vertebrae", (2000, 350))
    presets.append(ps)

    return presets



import logging
import time
import numpy as np

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel, QLineEdit, QPushButton, QDoubleSpinBox, QListWidget, \
    QHBoxLayout, QMenu, QAction, QComboBox

from opentps.core.data.plan._protonPlanDesign import ProtonPlanDesign
from opentps.core.data._patient import Patient
from opentps.core.data.plan._photonPlanDesign import PhotonPlanDesign
from opentps.core.io import mcsquareIO
from opentps.core.io.mcsquareIO import readBDL
from opentps.core.processing.doseCalculation.doseCalculationConfig import DoseCalculationConfig
from opentps.gui.panels.planDesignPanel.beamDialog import BeamDialog
from opentps.gui.panels.planDesignPanel.robustnessSettings import RobustnessSettings

logger = logging.getLogger(__name__)

class PlanDesignPanel(QWidget):
    def __init__(self, viewController):
        QWidget.__init__(self)

        self._patient:Patient = None
        self._viewController = viewController
        self._beamDescription = []

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self._modalityLabel = QLabel('Modality:')
        self.layout.addWidget(self._modalityLabel)
        self._modalityComboBox = QComboBox(self)
        self._modalityComboBox.addItem("IMPT")
        self._modalityComboBox.addItem("IMRT")
        self.layout.addWidget(self._modalityComboBox)
        
        self._planLabel = QLabel('Plan name:')
        self.layout.addWidget(self._planLabel)

        self._planIMPTNameEdit = QLineEdit(self)
        self._planIMPTNameEdit.setText('New IMPT plan design')
        self.layout.addWidget(self._planIMPTNameEdit)
        self._planIMRTNameEdit = QLineEdit(self)
        self._planIMRTNameEdit.setText('New IMRT plan design')
        self.layout.addWidget(self._planIMRTNameEdit)
        self._planIMRTNameEdit.hide()

        from opentps.gui.programSettingEditor import MCsquareConfigEditor
        self._mcsquareConfigWidget = MCsquareConfigEditor(self)
        self._mcsquareConfigWidget.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self._mcsquareConfigWidget)

        self._xBeamletSpacingLayout = QHBoxLayout()
        self.layout.addLayout(self._xBeamletSpacingLayout)
        self.xBeamletSpacingLabel = QLabel('Beamlet spacing X:')
        self._xBeamletSpacingLayout.addWidget(self.xBeamletSpacingLabel)
        self._xBeamletSpacing = QDoubleSpinBox()
        self._xBeamletSpacing.setGroupSeparatorShown(True)
        self._xBeamletSpacing.setRange(0.1, 100.0)
        self._xBeamletSpacing.setSingleStep(1.0)
        self._xBeamletSpacing.setValue(5.0)
        self._xBeamletSpacing.setSuffix(" mm")
        self._xBeamletSpacingLayout.addWidget(self._xBeamletSpacing)
        self._yBeamletSpacingLayout = QHBoxLayout()
        self.layout.addLayout(self._yBeamletSpacingLayout)
        self.yBeamletSpacingLabel = QLabel('Beamlet spacing Y:')
        self._yBeamletSpacingLayout.addWidget(self.yBeamletSpacingLabel)
        self._yBeamletSpacing = QDoubleSpinBox()
        self._yBeamletSpacing.setGroupSeparatorShown(True)
        self._yBeamletSpacing.setRange(0.1, 100.0)
        self._yBeamletSpacing.setSingleStep(1.0)
        self._yBeamletSpacing.setValue(5.0)
        self._yBeamletSpacing.setSuffix(" mm")
        self._yBeamletSpacingLayout.addWidget(self._yBeamletSpacing)

        self.xBeamletSpacingLabel.hide()
        self._xBeamletSpacing.hide()
        self.yBeamletSpacingLabel.hide()
        self._yBeamletSpacing.hide()

        self._spacingLayout = QHBoxLayout()
        self.layout.addLayout(self._spacingLayout)
        self._spacingLabel = QLabel('Spot spacing:')
        self._spacingLayout.addWidget(self._spacingLabel)
        self._spacingSpin = QDoubleSpinBox()
        self._spacingSpin.setGroupSeparatorShown(True)
        self._spacingSpin.setRange(0.1, 100.0)
        self._spacingSpin.setSingleStep(1.0)
        self._spacingSpin.setValue(5.0)
        self._spacingSpin.setSuffix(" mm")
        self._spacingLayout.addWidget(self._spacingSpin)

        self._layerLayout = QHBoxLayout()
        self.layout.addLayout(self._layerLayout)

        self._layerLabel = QLabel('Layer spacing:')
        self._layerLayout.addWidget(self._layerLabel)
        self._layerSpin = QDoubleSpinBox()
        self._layerSpin.setGroupSeparatorShown(True)
        self._layerSpin.setRange(0.1, 100.0)
        self._layerSpin.setSingleStep(1.0)
        self._layerSpin.setValue(2.0)
        self._layerSpin.setSuffix(" mm")
        self._layerLayout.addWidget(self._layerSpin)

        self._marginLayout = QHBoxLayout()
        self.layout.addLayout(self._marginLayout)

        self._marginLabel = QLabel('Target margin:')
        self._marginLayout.addWidget(self._marginLabel)
        self._marginSpin = QDoubleSpinBox()
        self._marginSpin.setGroupSeparatorShown(True)
        self._marginSpin.setRange(0.1, 100.0)
        self._marginSpin.setSingleStep(1.0)
        self._marginSpin.setValue(5.0)
        self._marginSpin.setSuffix(" mm")
        self._marginLayout.addWidget(self._marginSpin)

        self._proximalLayout = QHBoxLayout()
        self.layout.addLayout(self._proximalLayout)

        self._proximalLabel = QLabel('Proximal layers:')
        self._proximalLayout.addWidget(self._proximalLabel)
        self._proximalSpin = QDoubleSpinBox()
        self._proximalSpin.setGroupSeparatorShown(True)
        self._proximalSpin.setRange(0, 100)
        self._proximalSpin.setSingleStep(1)
        self._proximalSpin.setValue(1)
        self._proximalSpin.setDecimals(0)
        self._proximalLayout.addWidget(self._proximalSpin)

        self._distalLayout = QHBoxLayout()
        self.layout.addLayout(self._distalLayout)

        self._distalLabel = QLabel('Distal layers:')
        self._distalLayout.addWidget(self._distalLabel)
        self._distalSpin = QDoubleSpinBox()
        self._distalSpin.setGroupSeparatorShown(True)
        self._distalSpin.setRange(0, 1)
        self._distalSpin.setSingleStep(1)
        self._distalSpin.setValue(1)
        self._distalSpin.setDecimals(0)
        self._distalLayout.addWidget(self._distalSpin)

        self._isocenterLayout = QHBoxLayout()
        self.layout.addLayout(self._isocenterLayout)

        self._isocenterLabel = QLabel('Isocenter:')
        self._isocenterLayout.addWidget(self._isocenterLabel)
        self._isocenterComboBox = QComboBox(self)
        self._isocenterComboBox.addItem("Target COM")
        self._isocenterComboBox.addItem("Custom")
        self._isocenterLayout.addWidget(self._isocenterComboBox)
        self._isocenterCustomInput = QLineEdit(self)
        self._isocenterCustomInput.setPlaceholderText("(a,b,c) mm")
        self._isocenterCustomInput.hide()
        self._isocenterLayout.addWidget(self._isocenterCustomInput)
        self._isocenterComboBox.currentTextChanged.connect(self._toggleIsocenterInput)

        self._beams = QListWidget()
        self._beams.setContextMenuPolicy(Qt.CustomContextMenu)
        self._beams.customContextMenuRequested.connect(lambda pos, list_type='beam': self.List_RightClick(pos, list_type))
        self.layout.addWidget(self._beams)

        self._beamButton = QPushButton('Add beam')
        self.layout.addWidget(self._beamButton)
        self._beamButton.clicked.connect(self.add_new_beam)

        self._robustSettings = RobustnessSettings(self._viewController, parent=self)
        self.layout.addWidget(self._robustSettings)

        self._runButton = QPushButton('Design plan')
        self._runButton.clicked.connect(self._create)
        self.layout.addWidget(self._runButton)

        self.layout.addStretch()

        self.setCurrentPatient(self._viewController.currentPatient)
        self._modalityComboBox.currentTextChanged.connect(self.updateUIBasedOnModality)
        self._viewController.currentPatientChangedSignal.connect(self.setCurrentPatient)

    def _toggleIsocenterInput(self, text):
        if text == "Custom":
            self._isocenterCustomInput.show()
        else:
            self._isocenterCustomInput.hide()

    @property
    def selectedModality(self):
        return self._modalityComboBox.currentText()
    
    @selectedModality.setter
    def selectedModality(self, modality):
        self._modalityComboBox.setCurrentText(modality)
    
    def updateUIBasedOnModality(self,modality):
        if modality == "IMRT":
            self._mcsquareConfigWidget._txt2.hide()
            self._spacingLabel.hide()
            self._spacingSpin.hide()
            self._layerLabel.hide()
            self._layerSpin.hide()
            self._proximalLabel.hide()
            self._proximalSpin.hide()
            self._distalLabel.hide()
            self._distalSpin.hide()
            self._planIMPTNameEdit.hide()

            self.xBeamletSpacingLabel.show()
            self._xBeamletSpacing.show()
            self.yBeamletSpacingLabel.show()
            self._yBeamletSpacing.show()
            self._planIMRTNameEdit.show()

        elif modality == "IMPT":
            self._mcsquareConfigWidget._txt2.show()
            self._spacingLabel.show()
            self._spacingSpin.show()
            self._layerLabel.show()
            self._layerSpin.show()
            self._proximalLabel.show()
            self._proximalSpin.show()
            self._distalLabel.show()
            self._distalSpin.show()
            self._planIMPTNameEdit.show()

            self.xBeamletSpacingLabel.hide()
            self._xBeamletSpacing.hide()
            self.yBeamletSpacingLabel.hide()
            self._yBeamletSpacing.hide()
            self._planIMRTNameEdit.hide()

        for _ in range(self._beams.count()):
            self.delete_item('beam', 0)

        self._robustSettings._updateForModality(modality)

    def setCurrentPatient(self, patient:Patient):
        self._patient = patient

    def _create(self):
        logger.info('Plan is designed...')
        start = time.time()

        if self.selectedModality == "IMRT":
            planDesign = PhotonPlanDesign()
            planDesign.xBeamletSpacing_mm = self._xBeamletSpacing.value()
            planDesign.yBeamletSpacing_mm = self._yBeamletSpacing.value()
            planDesign.name = self._planIMRTNameEdit.text()

        elif self.selectedModality == "IMPT":
            planDesign = ProtonPlanDesign()
            planDesign.spotSpacing = self._spacingSpin.value()
            planDesign.layerSpacing = self._layerSpin.value()
            planDesign.name = self._planIMPTNameEdit.text()
            
        else:
            logger.error(f"Unsupported modality: {self.selectedModality}")

        planDesign.targetMargin = self._marginSpin.value()
        planDesign.patient = self._patient

        # Isocentre
        isocenter_option = self._isocenterComboBox.currentText()
        if isocenter_option == "Target COM":
            planDesign.isocenterPosition_mm = None
        elif isocenter_option == "Custom":
            try:
                custom_isocenter = eval(self._isocenterCustomInput.text())
                if isinstance(custom_isocenter, (list, tuple)) and len(custom_isocenter) == 3:
                    planDesign.isocenterPosition_mm = np.array(custom_isocenter)
                else:
                    raise ValueError("Invalid format for isocenter")
            except Exception as e:
                logger.error(f"Invalid isocenter input: {e}")
                return
        else:
            logger.error(f"Unsupported isocenter option: {isocenter_option}")

        # extract beam info from QListWidget
        beamNames = []
        gantryAngles = []
        couchAngles = []
        rangeShifters = []
        AlignLayers = False
        for i in range(self._beams.count()):
            BeamType = self._beamDescription[i]["BeamType"]
            if (BeamType == "beam"):
                beamNames.append(self._beamDescription[i]["BeamName"])
                gantryAngles.append(self._beamDescription[i]["GantryAngle"])
                couchAngles.append(self._beamDescription[i]["CouchAngle"])
                if self.selectedModality == "IMPT":
                    rs = self._beamDescription[i]["RangeShifter"]
                    rangeShifters.append(rs)

        planDesign.gantryAngles = gantryAngles
        planDesign.beamNames = beamNames
        planDesign.couchAngles = couchAngles
        if self.selectedModality == "IMPT": planDesign.rangeShifters = rangeShifters
        
        planDesign.robustness = self._robustSettings.robustness
        logger.info("New plan design created in {} sec".format(time.time() - start))

    def add_new_beam(self):
        beam_number = self._beams.count()

        # retrieve list of range shifters from BDL
        if self.selectedModality == "IMPT":
            bdl = readBDL(DoseCalculationConfig().bdlFile)
            RangeShifterList = [rs.ID for rs in bdl.rangeShifters]
        else: RangeShifterList=[]

        dialog = BeamDialog(self.selectedModality, "Beam " + str(beam_number + 1), RangeShifterList=RangeShifterList)
        if (dialog.exec()):
            BeamName = dialog.BeamName.text()
            GantryAngle = dialog.GantryAngle.value()
            CouchAngle = dialog.CouchAngle.value()
            if self.selectedModality == "IMPT":
                RangeShifter = dialog.RangeShifter.currentText()
                if (RangeShifter == "None"):
                    RS_disp = ""
                    rs = None
                else:
                    RS_disp = ", RS"
                    rs = [rsElem for rsElem in bdl.rangeShifters if rsElem.ID==RangeShifter]
                    if len(rs)==0:
                        rs = None
                    else:
                        rs = rs[0]
                self._beams.addItem(BeamName + ":  G=" + str(GantryAngle) + "°,  C=" + str(CouchAngle) + "°" + RS_disp)
                self._beamDescription.append(
                    {"BeamType": "beam", "BeamName": BeamName, "GantryAngle": GantryAngle, "CouchAngle": CouchAngle,
                    "RangeShifter": rs})
            else:
                self._beams.addItem(BeamName + ":  G=" + str(GantryAngle) + "°,  C=" + str(CouchAngle) + "°")
                self._beamDescription.append(
                    {"BeamType": "beam", "BeamName": BeamName, "GantryAngle": GantryAngle, "CouchAngle": CouchAngle})


    def List_RightClick(self, pos, list_type):
        if list_type == 'beam':
            item = self._beams.itemAt(pos)
            row = self._beams.row(item)
            pos = self._beams.mapToGlobal(pos)

        else:
            return

        if row > -1:
            self.context_menu = QMenu()
            self.delete_action = QAction("Delete")
            self.delete_action.triggered.connect(
                lambda checked, list_type=list_type, row=row: self.delete_item(list_type, row))
            self.context_menu.addAction(self.delete_action)
            self.context_menu.popup(pos)

    def delete_item(self, list_type, row):
        if list_type == 'beam':
            self._beams.takeItem(row)
            self._beamDescription.pop(row)




import functools
import inspect
import typing

from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QLineEdit, QPushButton, QMainWindow

from opentps.core.data.plan import PlanProtonBeam
from opentps.core.data.plan import PlanProtonLayer
from opentps.core.data.plan import PlanProtonSpot
from opentps.core.data.plan import RangeShifter
from opentps.core.data import PatientData
from opentps.core import Event

class PatientDataPropertyEditor(QWidget):
    def __init__(self, image, parent=None):
        super().__init__(parent=parent)

        self._mainLayout = QVBoxLayout(self)
        self.setLayout(self._mainLayout)

        for property in inspect.getmembers(image):
            # to remove private and protected
            # functions
            if not property[0].startswith('_'):
                # To remove other methods that
                # does not start with an underscore
                if not inspect.ismethod(property[1]):
                    if not isinstance(property[1], Event):
                        self._mainLayout.addWidget(TwoRowElement(property, parent=self))


class TwoRowElement(QWidget):
    def __init__(self, property, parent=None):
        super().__init__(parent)

        self._mainLayout = QHBoxLayout(self)
        self.setLayout(self._mainLayout)

        self._txt = QLabel(self)
        self._txt.setText(property[0])

        self._mainLayout.addWidget(self._txt)

        val = property[1]

        if isinstance(val, self.supportedTypes()):
            val = [val]

        if isinstance(val, typing.Sequence) and len(val)>0 and isinstance(val[0], self.supportedTypes()):
            for valElement in val:
                patientDataButton = QPushButton(self)
                patientDataButton.setText(str(valElement.__class__.__name__))
                patientDataButton.clicked.connect(functools.partial(self._openPatientData, valElement))
                self._mainLayout.addWidget(patientDataButton)
        else:
            self._nameLineEdit = QLineEdit(self)
            self._nameLineEdit.setText(str(val))
            self._txt.setBuddy(self._nameLineEdit)
            self._mainLayout.addWidget(self._nameLineEdit)

        self._mainLayout.addStretch()

    def _openPatientData(self, patientData:PatientData):
        w = QMainWindow(self.parent().parent())
        w.setWindowTitle('Image info')
        w.resize(400, 400)
        w.setCentralWidget(PatientDataPropertyEditor(patientData, self.parent().parent()))
        w.show()

    def supportedTypes(self):
        return typing.Union[PatientData, PlanProtonBeam, PlanProtonLayer, PlanProtonSpot, RangeShifter].__args__



from opentps.gui.main import run, patientList, viewController, runWithMainWindow

import opentps.gui.panels as panels
import opentps.gui.res as res
import opentps.gui.tools as tools
import opentps.gui.viewer as viewer
import opentps.gui.viewController as viewController

from opentps.core._loggingConfig import loggerConfig
loggerConfig().configure()



__all__ = [s for s in dir() if not s.startswith('_')]



from functools import partial
from typing import Optional, Union, Sequence

from vtkmodules import vtkImagingCore, vtkCommonCore
from vtkmodules.vtkFiltersCore import vtkContourFilter
from vtkmodules.vtkRenderingCore import vtkActor, vtkPolyDataMapper

from opentps.core.data.images._image3D import Image3D
from opentps.core.data.images import ROIMask
from opentps.core.data._roiContour import ROIContour
from opentps.gui.viewer.dataForViewer.ROIContourForViewer import ROIContourForViewer
from opentps.gui.viewer.dataForViewer.ROIMaskForViewer import ROIMaskForViewer


class ContourLayer:
    def __init__(self, renderer, renderWindow):
        self._contours = [] # Acts as a cache
        self._referenceImage = None
        self._renderer = renderer
        self._renderWindow = renderWindow
        self._resliceAxes = None
        self._vtkContours = []
        self._partialHandlers = []

    def close(self):
        for vtkContour in self._vtkContours:
            vtkContour.close()

    @property
    def contours(self) -> Sequence[Union[ROIContourForViewer, ROIMaskForViewer]]:
        return [contour for contour in self._contours]

    def setNewContour(self, contour:Union[ROIContour, ROIMask]):
        if isinstance(contour, ROIContour):
            contour = ROIContourForViewer(contour)
        elif isinstance(contour, ROIMask):
            contour = ROIMaskForViewer(contour)
        else:
            raise ValueError(str(type(contour)) + ' is not a valid type for a contour.')

        if contour in self._contours:
            return

        if not contour.visible:
            return

        contour.referenceImage = self.referenceImage

        self._contours.append(contour)

        vtkContourObj = vtkContour(contour, self._renderWindow)

        self._renderer.AddActor(vtkContourObj.actor)
        vtkContourObj.resliceAxes = self._resliceAxes
        self._vtkContours.append(vtkContourObj)

        partialHandler = partial(self._handleVisibilityChange, contour)
        self._partialHandlers.append(partialHandler)
        contour.visibleChangedSignal.connect(partialHandler)

        self._renderWindow.Render()

    def _handleVisibilityChange(self, contour:Union[ROIContourForViewer, ROIMaskForViewer], visible):
        if not contour.visible:
            self._removeContour(contour)

    def _removeContour(self, contour:Union[ROIContourForViewer, ROIMaskForViewer]):
        contourIndex = self._contours.index(contour)
        vtkContourObj = self._vtkContours[contourIndex]
        partialHandler = self._partialHandlers[contourIndex]
        self._renderer.RemoveActor(vtkContourObj.actor)

        self._contours.remove(contour)
        self._vtkContours.remove(vtkContourObj)
        self._partialHandlers.remove(partialHandler)

        contour.visibleChangedSignal.disconnect(partialHandler)


    @property
    def referenceImage(self) -> Optional[Image3D]:
        return self._referenceImage

    @referenceImage.setter
    def referenceImage(self, image: Image3D):
        self._referenceImage = image

        for contour in self._contours:
            contour.referenceImage = self._referenceImage

    @property
    def resliceAxes(self):
        return self._resliceAxes

    @resliceAxes.setter
    def resliceAxes(self, resliceAxes):
        self._resliceAxes = resliceAxes
        self._updateContoursResliceAxes()

    def _updateContoursResliceAxes(self):
        for vtkContour in self._vtkContours:
            vtkContour.resliceAxes = self._resliceAxes

    def resliceDataFromPhysicalPoint(self, point):
        return [contour.resliceDataFromPhysicalPoint(point) for contour in self._vtkContours]

class vtkContour:
    def __init__(self, contour, renderWindow):
        self.actor = vtkActor()
        self._contour = contour
        self.contourFilter = vtkContourFilter()
        self.mapper = vtkPolyDataMapper()
        self.renderWindow = renderWindow  # Not very beautiful to pass renderWindow but fewer lines of code than trigering a render event
        self.reslice = vtkImagingCore.vtkImageReslice()

        self.actor.SetMapper(self.mapper)

        self.reslice.SetOutputDimensionality(2)
        self.reslice.SetInterpolationModeToNearestNeighbor()

        self.reslice.SetInputConnection(self._contour.vtkOutputPort)
        self.contourFilter.SetInputConnection(self.reslice.GetOutputPort())
        self.mapper.SetInputConnection(self.contourFilter.GetOutputPort())

        self.contourFilter.SetValue(0, 0.1)
        self.contourFilter.Update()

        self.actor.GetProperty().SetLineWidth(3)

        self.reloadColor()

        self.setVisible(self._contour.visible)

        self._contour.visibleChangedSignal.connect(self.setVisible)
        self._contour.colorChangedSignal.connect(self.reloadColor)

    def close(self):
        self._contour.visibleChangedSignal.disconnect(self.setVisible)
        self._contour.colorChangedSignal.disconnect(self.reloadColor)

    @property
    def resliceAxes(self):
        return self.reslice.GetResliceAxes()

    @resliceAxes.setter
    def resliceAxes(self, resliceAxes):
        self.reslice.SetResliceAxes(resliceAxes)

    def reloadColor(self):
        imageColor = self._contour.color

        # Create a greyscale lookup table
        table = vtkCommonCore.vtkLookupTable()
        table.SetRange(0, 1)  # image intensity range
        table.SetValueRange(0.0, 1.0)  # from black to white
        table.SetSaturationRange(0.0, 0.0)  # no color saturation
        table.SetRampToLinear()

        table.SetNumberOfTableValues(2)
        table.SetTableValue(0, (imageColor[0] / 255.0, imageColor[1] / 255.0, imageColor[2] / 255.0, 1))
        table.SetTableValue(1, (imageColor[0] / 255.0, imageColor[1] / 255.0, imageColor[2] / 255.0, 1))
        table.SetBelowRangeColor(0, 0, 0, 0)
        table.SetUseBelowRangeColor(True)
        table.SetAboveRangeColor(imageColor[0] / 255.0, imageColor[1] / 255.0, imageColor[2] / 255.0, 1)
        table.SetUseAboveRangeColor(True)
        table.Build()

        # contourActor.GetProperty().SetColor(imageColor[0], imageColor[1], imageColor[2])
        self.mapper.SetLookupTable(table)

    def setVisible(self, visible: bool):
        self.actor.SetVisibility(visible)
        self.renderWindow.Render()

    def resliceDataFromPhysicalPoint(self, point):
        imageData = self.reslice.GetInput(0)

        ind = [0, 0, 0]
        imageData.TransformPhysicalPointToContinuousIndex(point, ind)
        return imageData.GetScalarComponentAsFloat(int(ind[0]), int(ind[1]), int(ind[2]), 0)



from PyQt5.QtGui import QImage, QPixmap, QPainter, QPalette, QBrush, QPolygon
from PyQt5.QtCore import QPoint
import numpy as np
import math


from PyQt5.QtWidgets import *
from PyQt5.QtCore import Qt, QDir#, pyqtSignal
from opentps.core import createDRRDynamic2DSequences



class DRRPanel(QWidget):

    # fluoroSeqCreated = pyqtSignal()

    def __init__(self, viewController):
        QWidget.__init__(self)

        self._viewController = viewController
        self._viewController.patientAddedSignal.connect(self.refreshDataList)

        self.patients = self._viewController.patientList
        # self.toolbox_width = toolbox_width
        self.data_path = QDir.currentPath()
        self.CT_disp_ID = ""

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self.imageSelectionBox = QComboBox()
        # self.refreshDataList(self._viewController.currentPatient)

        self.layout.addWidget(self.imageSelectionBox) ## --> this must changed to a data selection box (self.imageSelectionBox)


        self.fluoSimButton = QPushButton('Fluoroscopy simulation')
        self.layout.addWidget(self.fluoSimButton)
        self.fluoSimButton.clicked.connect(self.selectProjectionAngles)

    def refreshDataList(self, patient):

        print('in DRRPanel refresh data list ')
        for data in patient.getPatientDataOfType('CTImage'):
            name = data.name
            self.imageSelectionBox.addItem(name, data)
        for data in patient.getPatientDataOfType('Dynamic3DSequence'):
            name = data.name
            self.imageSelectionBox.addItem(name, data)

    """
    class PatientComboBox(QComboBox):
    def __init__(self, viewController):
        QComboBox.__init__(self)

        self._viewController = viewController

        self._viewController.patientAddedSignal.connect(self._addPatient)
        self._viewController.patientRemovedSignal.connect(self._removePatient)

        self.currentIndexChanged.connect(self._setCurrentPatient)

    def _addPatient(self, patient):
        name = patient.name
        if name is None:
            name = 'None'

        self.addItem(name, patient)
        if self.count() == 1:
            self._viewController.currentPatient = patient

    def _removePatient(self, patient):
        self.removeItem(self.findData(patient))

    def _setCurrentPatient(self, index):
        self._viewController.currentPatient = self.currentData()
    """

    def Data_path_changed(self, data_path):
        self.data_path = data_path

    def selectProjectionAngles(self):

        fluoroAngleDialog = ChoseAngles_dialog(self._viewController.currentPatient.dynamic3DSequences[0].dyn3DImageList[0])

        if (fluoroAngleDialog.exec()):

            anglesAndAxisList = fluoroAngleDialog.anglesList
            print('in selectProjectionAngles', anglesAndAxisList)
            print('this should be change to use the currently selected patient and dyn seq if several')

            # fluoroSeq = self.createDRRs(angles, self._viewController.currentPatient.dynamic3DSequences[0].dyn3DImageList, self._viewController.currentPatient)

            print('in drrPanel selectProjectionAngles !!! source image is hard coded for now')
            sourceImageOrSeq = self._viewController.currentPatient.dynamic3DSequences[0]
            dyn2DDRRSeqList = createDRRDynamic2DSequences(sourceImageOrSeq, anglesAndAxisList)

            # plt.figure()
            # plt.imshow(dyn2DDRRSeqList[0].dyn2DImageList[0].imageArray)
            # plt.show()

            if sourceImageOrSeq.patient != None:
                for dyn2DSeq in dyn2DDRRSeqList:
                    sourceImageOrSeq.patient.appendPatientData(dyn2DSeq)

            # self.fluoroSeqCreated.emit()
            #self.data_path = QFileDialog.getExistingDirectory(self, "Select 4D data folder", self.data_path)
            #self.Patients.importDyn4DSeq(self.data_path)

            # display Dynamic series
#             for series in self.Patients.list[0].Dyn2DSeqList:
#                 if (series.isLoaded == 1):
#                     serieRoot = StandardItem(txt=series.SequenceName)
#                     for image in series.dyn2DImageList:
#                         imageName = StandardItem(txt=image.ImgName)
#                         serieRoot.appendRow(imageName)
#                     self.rootNode.appendRow(serieRoot)
#                     self.treeView.setModel(self.treeModel)
#                     self.treeView.expandAll()
#                     self.treeView.clicked.connect(self.getTreeViewValue)
#
    # def createDRRs(self, angleList, Dyn4DSeq, patient, orientation='Axial'):
    #
    #     try:
    #         import tomopy
    #
    #         for anglesAndOri in angleList:
    #
    #             angle = anglesAndOri[0]
    #             orientation = anglesAndOri[1]
    #
    #             fluoroSeq = Dynamic2DSequence()
    #             fluoroSeq.type = 'Fluoroscopy simulation'
    #             fluoroSeq.projectionAngle = angle
    #             fluoroSeq.SequenceName = 'FluoSim_' + anglesAndOri[1] + '_' + str(
    #                 int(np.round(angle * 360 / (2 * math.pi))))
    #             for imageIndex, image in enumerate(Dyn4DSeq.dyn3DImageList):
    #
    #                 # to rotate around Z axis
    #                 if orientation == 'Axial':
    #                     imageToUse = image.Image.transpose(2, 1, 0)
    #                 if orientation == 'Coronal':
    #                     imageToUse = image.Image.transpose(1, 2, 0)
    #                 if orientation == 'Sagittal':
    #                     imageToUse = image.Image.transpose(0, 2, 1)
    #
    #                 print('! ! ! in fluoroscopy_sim --> les orientation et transpose ici sont à vérifier! ! !')
    #                 fluoSimImage = tomopy.project(imageToUse, angle)[0]
    #
    #                 # plt.figure()
    #                 # plt.imshow(fluoSimImage)
    #                 # plt.show()
    #
    #                 image_2D = image2D()
    #                 image_2D.Image = fluoSimImage
    #                 image_2D.ImgName = str(imageIndex + 1)
    #                 image_2D.PixelSpacing = [1, 1]
    #                 fluoroSeq.dyn2DImageList.append(image_2D)
    #                 fluoroSeq.isLoaded = True
    #
    #             patient.Dyn2DSeqList.append(fluoroSeq)
    #
    #     except:
    #         print("No module tomopy available")
# class StandardItem(QStandardItem):
#   def __init__(self, txt="", fontSize = 12, setBold=False, color=QColor(0,0,0)):
#     super().__init__()
#     self.setEditable(False)
#     self.setForeground(color)
#     self.setText(txt)



class ChoseAngles_dialog(QDialog):

    def __init__(self, Image3D):

        # initialize the window
        QDialog.__init__(self)

        self.main_layout = QGridLayout()
        self.setLayout(self.main_layout)

        self.infoLabel = QLabel()
        self.infoLabel.viewer_palette = QPalette()
        self.infoLabel.viewer_palette.setColor(QPalette.Window, Qt.black)
        self.infoLabel.setAlignment(Qt.AlignTop)
        self.infoLabel.setStyleSheet('color: black')
        self.infoLabelText = ''
        self.main_layout.addWidget(self.infoLabel, 0, 0)

        self.imageLabel = QLabel()
        self.imageLabel.setMaximumSize(512, 512)
        self.imageLabel.setAlignment(Qt.AlignTop)
        self.main_layout.addWidget(self.imageLabel, 1, 0)

        self.orientation = "Z"
        self.image3D = self.prepare_image_for_viewer(Image3D.imageArray)
        self.imageShape = self.image3D.shape
        self.resolution = Image3D.spacing
        self.imageCenter = [self.imageLabel.pos().x() + int(self.imageLabel.width() / 2), self.imageLabel.pos().y() + int(self.imageLabel.height() / 2)]
        print('in ChoseAngles_dialog constructor', self.imageLabel.pos().x())
        print(self.imageCenter)
        self.distanceFromSourceToCenter = min(self.imageCenter[0], self.imageCenter[1]) - 10
        self.pannelWidth = min(self.imageLabel.width(), self.imageLabel.height())

        self.anglesList = []

        self.currentMousePos = QPoint(0, 0)
        self.angleBetweenMouseAndCenter = 0
        self.xRaySourcePoint = QPoint(0, 0)
        self.pannelPoint1 = QPoint(0, 0)
        self.pannelPoint2 = QPoint(0, 0)
        self.getPolygonPoints(self.angleBetweenMouseAndCenter)
        self.colorList = [Qt.cyan, Qt.red, Qt.blue]
        self.colorList *= 5

        # buttons
        self.ButtonLayout = QGridLayout()
        self.AddAngleButton = QPushButton('Add angle')
        self.AddAngleButton.clicked.connect(self.addAngle)
        self.ButtonLayout.addWidget(self.AddAngleButton, 1, 0)
        self.ValidateButton = QPushButton('OK')
        self.ValidateButton.clicked.connect(self.accept)
        self.ButtonLayout.addWidget(self.ValidateButton, 1, 1)
        self.CancelButton = QPushButton('Cancel')
        self.ButtonLayout.addWidget(self.CancelButton, 1, 2)
        self.CancelButton.clicked.connect(self.reject)
        self.XYButton = QPushButton('Z')
        self.XYButton.clicked.connect(self.changeImageView)
        self.ButtonLayout.addWidget(self.XYButton, 0, 0)
        self.YZButton = QPushButton('Y')
        self.YZButton.clicked.connect(self.changeImageView)
        self.ButtonLayout.addWidget(self.YZButton, 0, 1)
        self.XZButton = QPushButton('X')
        self.XZButton.clicked.connect(self.changeImageView)
        self.ButtonLayout.addWidget(self.XZButton, 0, 2)

        self.main_layout.addLayout(self.ButtonLayout, 2, 0)

        self.setWindowTitle('Select angles to use for fluoroscopy projections')

        self.selectSlice()
        self.resize(512, 512)
        self.update_viewer()

##-----------------------------------------------------------------------------------------------------------
    def selectSlice(self):

        if self.orientation == 'Z':
            self.current_sliceIndex = round(self.imageShape[2] / 2)
            self.sliceToShow = self.image3D[:, :, self.current_sliceIndex].transpose(1,0)
        elif self.orientation == 'X':
            self.current_sliceIndex = round(self.imageShape[0] / 2)
            self.sliceToShow = np.flip(self.image3D[self.current_sliceIndex, :, :].transpose(1,0), 0)
        elif self.orientation == 'Y':
            self.current_sliceIndex = round(self.imageShape[1] / 2)
            self.sliceToShow = np.flip(self.image3D[:, self.current_sliceIndex, :].transpose(1,0), 0)

        self.sliceToShow = np.require(self.sliceToShow, np.uint8, 'C')


##-----------------------------------------------------------------------------------------------------------
    def getPolygonPoints(self, angle):

        print('in get polygon points')
        print(self.imageCenter)
        print(self.imageCenter[0] - 2 * (self.distanceFromSourceToCenter * math.sin(angle)))
        print(self.imageCenter[1] - 2 * (self.distanceFromSourceToCenter * math.cos(angle)))

        self.xRaySourcePoint = QPoint(int(self.imageCenter[0] - 2 * (self.distanceFromSourceToCenter * math.sin(angle))),
                                      int(self.imageCenter[1] - 2 * (self.distanceFromSourceToCenter * math.cos(angle))))
        self.pannelPoint1 = QPoint(int(self.imageCenter[0] + (self.distanceFromSourceToCenter * math.sin(angle)) + math.cos(angle)*self.pannelWidth/2),
                                    int(self.imageCenter[1] + (self.distanceFromSourceToCenter * math.cos(angle)) - math.sin(angle)*self.pannelWidth/2))
        self.pannelPoint2 = QPoint(int(self.imageCenter[0] + (self.distanceFromSourceToCenter * math.sin(angle)) - math.cos(angle)*self.pannelWidth/2),
                                      int(self.imageCenter[1] + (self.distanceFromSourceToCenter * math.cos(angle)) + math.sin(angle)*self.pannelWidth/2))

##-----------------------------------------------------------------------------------------------------------
    def update_viewer(self):

        # calculate scaling factor
        if self.orientation == 'Z':
            Yscaling = self.resolution[1] / self.resolution[0]
        elif self.orientation == 'X':
            Yscaling = self.resolution[2] / self.resolution[1]
        elif self.orientation == 'Y':
            Yscaling = self.resolution[2] / self.resolution[0]

        self.imageCenter = [self.imageLabel.pos()._x() + int(self.imageLabel.width() / 2),
                            self.imageLabel.pos()._y() + int(self.imageLabel.height() / 2)]
        self.distanceFromSourceToCenter = min(self.imageCenter[0], self.imageCenter[1]) - 10
        self.pannelWidth = min(self.imageLabel.width(), self.imageLabel.height())

        imageQT = QImage(self.sliceToShow, self.sliceToShow.shape[1], self.sliceToShow.shape[0], self.sliceToShow.shape[1], QImage.Format_Indexed8)
        imageQT = imageQT.convertToFormat(QImage.Format_ARGB32)
        MergedImage = QImage(self.sliceToShow.shape[1], round(self.sliceToShow.shape[0] * Yscaling), QImage.Format_ARGB32)
        MergedImage.fill(Qt.black)
        painter = QPainter()
        painter.begin(MergedImage)
        painter.scale(1.0, Yscaling)
        painter.drawImage(0, 0, imageQT)

        for angleIndex, angle in enumerate(self.anglesList):
            self.draw_polygon(angle[0], angleIndex, painter)
        self.draw_polygon(self.angleBetweenMouseAndCenter, -1, painter)

        painter.end()

        self.imageLabel.setPixmap(QPixmap.fromImage(MergedImage).scaledToWidth(self.width() - 10, mode=Qt.SmoothTransformation))

##-----------------------------------------------------------------------------------------------------------
    def addAngle(self):
        self.anglesList.append([self.angleBetweenMouseAndCenter, self.orientation])
        self.infoLabelText += 'Angle ' + str(len(self.anglesList)) + ': ' + str(int(np.round(self.anglesList[-1][0]*(360/(2*math.pi))))) + '°' + '\n'
        self.infoLabel.setText(self.infoLabelText)

##-----------------------------------------------------------------------------------------------------------
    def draw_polygon(self, angle, angleIndex, painter):

        self.getPolygonPoints(angle)
        if angleIndex == -1:
            painter.setBrush(QBrush(Qt.yellow, Qt.Dense6Pattern))
        else:
            painter.setBrush(QBrush(self.colorList[angleIndex], Qt.Dense6Pattern))

        polygon = QPolygon([self.pannelPoint1,
                            self.pannelPoint2,
                            self.xRaySourcePoint])

        painter.drawPolygon(polygon)

    # def onClicked(self, item):
    #     print('test onClicked')

    # def onDoubleClick(self, item):
    #     print('test onDoubleClick')

##-----------------------------------------------------------------------------------------------------------
    def mouseMoveEvent(self, QMouseEvent):
        #print('Mouse coords: ( %d : %d )' % (QMouseEvent.x(), QMouseEvent.y()))
        self.currentMousePos.setX(QMouseEvent._x())
        self.currentMousePos.setY(QMouseEvent._y())
        self.getAngleFromMousePosition()

        self.infoLabel.setText(self.infoLabelText + "Current angle : " + str(int(np.round(self.angleBetweenMouseAndCenter*(360/6.28)))) + '°')
        self.update_viewer()

##-----------------------------------------------------------------------------------------------------------
    def getAngleFromMousePosition(self):

        if self.imageCenter[1] == self.currentMousePos.y() and self.imageCenter[0] > self.currentMousePos.x():
            self.angleBetweenMouseAndCenter = math.pi/2
        elif self.imageCenter[1] == self.currentMousePos.y() and self.imageCenter[0] < self.currentMousePos.x():
            self.angleBetweenMouseAndCenter = 3 * math.pi / 2
        elif self.imageCenter[1] < self.currentMousePos.y() and self.imageCenter[0] == self.currentMousePos.x():
            self.angleBetweenMouseAndCenter = math.pi
        elif self.imageCenter[1] > self.currentMousePos.y() and self.imageCenter[0] == self.currentMousePos.x():
            self.angleBetweenMouseAndCenter = 0
        elif self.imageCenter[0] > self.currentMousePos.x() and self.imageCenter[1] > self.currentMousePos.y():
            self.angleBetweenMouseAndCenter = math.atan(abs(self.imageCenter[0]-self.currentMousePos.x())/abs(self.imageCenter[1]-self.currentMousePos.y()))
        elif self.imageCenter[0] > self.currentMousePos.x() and self.imageCenter[1] < self.currentMousePos.y():
            self.angleBetweenMouseAndCenter = math.pi - math.atan(abs(self.imageCenter[0]-self.currentMousePos.x())/abs(self.imageCenter[1]-self.currentMousePos.y()))
        elif self.imageCenter[0] < self.currentMousePos.x() and self.imageCenter[1] < self.currentMousePos.y():
            self.angleBetweenMouseAndCenter = math.pi + math.atan(abs(self.imageCenter[0]-self.currentMousePos.x())/abs(self.imageCenter[1]-self.currentMousePos.y()))
        elif self.imageCenter[0] < self.currentMousePos.x() and self.imageCenter[1] > self.currentMousePos.y():
            self.angleBetweenMouseAndCenter = (2 * math.pi) - math.atan(abs(self.imageCenter[0]-self.currentMousePos.x())/abs(self.imageCenter[1]-self.currentMousePos.y()))

    ##-----------------------------------------------------------------------------------------------------------
    def changeImageView(self):
        self.orientation = self.sender().text()
        self.selectSlice()
        self.update_viewer()

    ##-----------------------------------------------------------------------------------------------------------
    def prepare_image_for_viewer(self, image):
        img_min = image.min()
        # img_max = self.Image.max()
        img_max = np.percentile(image, 99.995)
        img = 255 * (image.astype(np.float32) - img_min) / (
                    img_max - img_min + 1e-5)  # normalize data betwee, 0 and 255
        img[img > 255] = 255
        return img


from PyQt5.QtWidgets import *
from PyQt5.QtGui import QColor, QPixmap, QIcon

from opentps.core.data.images import ROIMask
from opentps.core.data._patient import Patient
from opentps.core.data._roiContour import ROIContour
from opentps.core.data._rtStruct import RTStruct
from opentps.gui.viewer.dataForViewer.ROIContourForViewer import ROIContourForViewer
from opentps.gui.viewer.dataForViewer.ROIMaskForViewer import ROIMaskForViewer


class ROIPanel(QWidget):
  def __init__(self, viewController):
    QWidget.__init__(self)

    self.items = []
    self.layout = QVBoxLayout()
    self._patient = None
    self._viewController = viewController

    self.setLayout(self.layout)

    self._filterEdit = QLineEdit(self)
    self._filterEdit.setPlaceholderText('Filter')
    self._filterEdit.textEdited.connect(self._setFilteredROIs)
    self.layout.addWidget(self._filterEdit)

    self._listFrame = QFrame(self)
    self.layout.addWidget(self._listFrame)
    self._listLayout = QVBoxLayout()
    self._listFrame.setLayout(self._listLayout)

    self.layout.addStretch()

    self.setCurrentPatient(self._viewController.currentPatient)
    self._viewController.currentPatientChangedSignal.connect(self.setCurrentPatient)

  def setCurrentPatient(self, patient:Patient):
    if patient==self._patient:
      return
    elif not self._patient is None:
      self._patient.rtStructAddedSignal.disconnect(self._setFilteredROIs)
      self._patient.rtStructRemovedSignal.disconnect(self._setFilteredROIs)
      self._patient.roiMaskAddedSignal.disconnect(self._setFilteredROIs)
      self._patient.roiMaskRemovedSignal.disconnect(self._setFilteredROIs)

      self._resetList()

    self._patient = patient

    self._setFilteredROIs()

    if not (self._patient is None):
      self._patient.rtStructAddedSignal.connect(self._setFilteredROIs)
      self._patient.rtStructRemovedSignal.connect(self._setFilteredROIs)
      self._patient.roiMaskAddedSignal.connect(self._setFilteredROIs)
      self._patient.roiMaskRemovedSignal.connect(self._setFilteredROIs)

  def _resetList(self):
    for widget in self.items:
      if isinstance(widget, ROIItem):
        widget.setVisible(False)
        self._listLayout.removeWidget(widget)
    self.items = []

  def _setFilteredROIs(self, *args):
    if self._patient is None:
      return

    enteredText = self._filterEdit.text().lower()

    self._resetList()

    for rtStruct in self._patient.rtStructs:
      for contour in rtStruct.contours:
        if enteredText in contour.name.lower() or enteredText=='':
          self.addROIContour(contour)

    for roiMask in self._patient.roiMasks:
      if enteredText in roiMask.name.lower() or enteredText=='':
        self.addROIMask(roiMask)

  def addROIContour(self, contour:ROIContour):
      checkbox = ROIItem(ROIContourForViewer(contour), self._viewController)

      self._listLayout.addWidget(checkbox)
      self.items.append(checkbox)

  def addROIMask(self, roiMask:ROIMask):
    checkbox = ROIItem(ROIMaskForViewer(roiMask), self._viewController)

    self._listLayout.addWidget(checkbox)
    self.items.append(checkbox)

  def removeRTStruct(self, rtStruct:RTStruct):
    for contour in rtStruct.contours:
      for item in self.items:
        if item._contour == contour:
          self._listLayout.removeWidget(item)
          item.setParent(None)
          return

  def removeROIMask(self, roiMask:ROIMask):
    for item in self.items:
        if item._contour == roiMask:
          self._listLayout.removeWidget(item)
          item.setParent(None)
          return

class ROIItem(QCheckBox):
  def __init__(self, contour, viewController):
    super().__init__(contour.name)

    self._contour = contour
    self._viewController = viewController

    self.setChecked(self._contour.visible)

    self._contour.visibleChangedSignal.connect(self.setChecked)

    self.clicked.connect(lambda c: self.handleClick(c))

    pixmap = QPixmap(100, 100)
    pixmap.fill(QColor(contour.color[0], contour.color[1], contour.color[2], 255))
    self.setIcon(QIcon(pixmap))

  @property
  def contour(self):
    return self._contour

  def handleClick(self, isChecked):
    self._contour.visible = isChecked
    self._viewController.showContour(self._contour.data)



from typing import Optional

from vtkmodules import vtkCommonMath
from vtkmodules.vtkCommonColor import vtkNamedColors
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import vtkPolyLine, vtkCellArray, vtkPolyData
from vtkmodules.vtkFiltersSources import vtkSphereSource, vtkLineSource
from vtkmodules.vtkRenderingCore import vtkActor, vtkPolyDataMapper

from opentps.core.data.images._image3D import Image3D
from opentps.core.data.plan import PlanProtonBeam
from opentps.core.data.plan._rtPlan import RTPlan
from opentps.core.processing.imageProcessing import imageTransform3D


class BeamLayer:
    def __init__(self, renderer, renderWindow):
        self._renderer = renderer
        self._renderWindow = renderWindow
        self._resliceAxes = None

        self._sphereSource = vtkSphereSource()
        self._sphereSource.SetCenter(0.0, 0.0, 0.0)
        self._sphereSource.SetRadius(5.0)

        mapper = vtkPolyDataMapper()
        mapper.SetInputConnection(self._sphereSource.GetOutputPort())

        colors = vtkNamedColors()

        self._sphereActor = vtkActor()
        self._sphereActor.SetMapper(mapper)
        self._sphereActor.GetProperty().SetColor(colors.GetColor3d("Fuchsia"))
        self._sphereActor.SetVisibility(False)

        renderer.AddActor(self._sphereActor)

        p0 = [0.0, 0.0, 0.0]
        p1 = [0.0, 0.0, 0.0]

        self._lineSource = vtkLineSource()
        self._lineSource.SetPoint1(p0)
        self._lineSource.SetPoint2(p1)

        self._lineMapper = vtkPolyDataMapper()

        self._lineActor = vtkActor()
        self._lineActor.SetMapper(self._lineMapper)
        self._lineActor.GetProperty().SetColor(colors.GetColor3d("Fuchsia"))
        self._lineActor.SetVisibility(False)

        renderer.AddActor(self._lineActor)

    def close(self):
        self._sphereActor.SetVisibility(False)
        self._lineActor.SetVisibility(False)

        self._renderer.RemoveActor(self._sphereActor)
        self._renderer.RemoveActor(self._lineActor)

    @property
    def resliceAxes(self):
        return self._resliceAxes

    @resliceAxes.setter
    def resliceAxes(self, resliceAxes):
        self._resliceAxes = resliceAxes

        # TODO: update view

    def setBeam(self, beam:PlanProtonBeam, referenceImage:Image3D):
        transfo_mat = vtkCommonMath.vtkMatrix4x4()
        transfo_mat.DeepCopy(self._resliceAxes)
        transfo_mat.Invert()
        posAfterInverse = transfo_mat.MultiplyPoint((beam.isocenterPosition[0], beam.isocenterPosition[1], beam.isocenterPosition[2], 1))

        self._sphereSource.SetCenter(posAfterInverse[0], posAfterInverse[1], 0)

        point2 = imageTransform3D.dicomCoordinate2iecGantry(beam, beam.isocenterPosition)
        point2 = imageTransform3D.iecGantryCoordinatetoDicom(beam, [point2[0], point2[1], point2[2] - 500])

        posAfterInverse2 = transfo_mat.MultiplyPoint((point2[0], point2[1], point2[2], 1))

        points = vtkPoints()
        points.InsertNextPoint((posAfterInverse[0], posAfterInverse[1], 0.01))
        points.InsertNextPoint((posAfterInverse2[0], posAfterInverse2[1], 0.01))

        polyLine = vtkPolyLine()
        polyLine.GetPointIds().SetNumberOfIds(2)
        for i in range(0, 2):
            polyLine.GetPointIds().SetId(i, i)

        cells = vtkCellArray()
        cells.InsertNextCell(polyLine)

        polyData = vtkPolyData()
        polyData.SetPoints(points)
        polyData.SetLines(cells)

        self._lineMapper.SetInputData(polyData)

        self._sphereActor.SetVisibility(True)
        self._lineActor.SetVisibility(True)

        self._renderWindow.Render()


class RTPlanLayer:
    def __init__(self, renderer, renderWindow):
        self._renderer = renderer
        self._renderWindow = renderWindow
        self._resliceAxes = None
        self._plan = None

        self._beamLayers = []

    def close(self):
        for bLayer in self._beamLayers:
            bLayer.close()

        self._beamLayers = []

        self._renderWindow.Render()

    @property
    def resliceAxes(self):
        return self._resliceAxes

    @resliceAxes.setter
    def resliceAxes(self, resliceAxes):
        self._resliceAxes = resliceAxes

        for bl in self._beamLayers:
            bl.resliceAxes = self._resliceAxes

    @property
    def plan(self) -> Optional[RTPlan]:
        return self._plan

    def setPlan(self, plan:RTPlan, referenceImage:Image3D):
        if plan is None:
            self._plan = None
            self.close()
            return
        elif self._plan == plan:
            return

        self._plan = plan

        self.close()

        for beam in plan:
            bLayer = BeamLayer(self._renderer, self._renderWindow)
            bLayer.resliceAxes = self._resliceAxes
            bLayer.setBeam(beam, referenceImage)
            self._beamLayers.append(bLayer)

        self._renderWindow.Render()